/****************************************************************************
** Meta object code from reading C++ file 'ModelDataTableDialog.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/widgets/ModelDataTableDialog.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'ModelDataTableDialog.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv20ModelDataTableDialogE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::ModelDataTableDialog::qt_create_metaobjectdata<qt_meta_tag_ZN2sv20ModelDataTableDialogE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::ModelDataTableDialog",
        "scrollToFrame",
        "",
        "sv_frame_t",
        "frame",
        "userScrolledToFrame",
        "playbackScrolledToFrame",
        "addCommand",
        "Command*",
        "viewClicked",
        "QModelIndex",
        "viewPressed",
        "currentChanged",
        "currentChangedThroughResort",
        "searchTextChanged",
        "searchRepeated",
        "insertRow",
        "deleteRows",
        "editRow",
        "togglePlayTracking",
        "modelRemoved"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'scrollToFrame'
        QtMocHelpers::SignalData<void(sv_frame_t)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Slot 'userScrolledToFrame'
        QtMocHelpers::SlotData<void(sv_frame_t)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Slot 'playbackScrolledToFrame'
        QtMocHelpers::SlotData<void(sv_frame_t)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Slot 'addCommand'
        QtMocHelpers::SlotData<void(Command *)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 2 },
        }}),
        // Slot 'viewClicked'
        QtMocHelpers::SlotData<void(const QModelIndex &)>(9, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 10, 2 },
        }}),
        // Slot 'viewPressed'
        QtMocHelpers::SlotData<void(const QModelIndex &)>(11, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 10, 2 },
        }}),
        // Slot 'currentChanged'
        QtMocHelpers::SlotData<void(const QModelIndex &, const QModelIndex &)>(12, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 10, 2 }, { 0x80000000 | 10, 2 },
        }}),
        // Slot 'currentChangedThroughResort'
        QtMocHelpers::SlotData<void(const QModelIndex &)>(13, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 10, 2 },
        }}),
        // Slot 'searchTextChanged'
        QtMocHelpers::SlotData<void(const QString &)>(14, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'searchRepeated'
        QtMocHelpers::SlotData<void()>(15, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'insertRow'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'deleteRows'
        QtMocHelpers::SlotData<void()>(17, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'editRow'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'togglePlayTracking'
        QtMocHelpers::SlotData<void()>(19, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'modelRemoved'
        QtMocHelpers::SlotData<void()>(20, 2, QMC::AccessProtected, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ModelDataTableDialog, qt_meta_tag_ZN2sv20ModelDataTableDialogE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::ModelDataTableDialog::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv20ModelDataTableDialogE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv20ModelDataTableDialogE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv20ModelDataTableDialogE_t>.metaTypes,
    nullptr
} };

void sv::ModelDataTableDialog::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ModelDataTableDialog *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->scrollToFrame((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 1: _t->userScrolledToFrame((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 2: _t->playbackScrolledToFrame((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 3: _t->addCommand((*reinterpret_cast<std::add_pointer_t<Command*>>(_a[1]))); break;
        case 4: _t->viewClicked((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 5: _t->viewPressed((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 6: _t->currentChanged((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[2]))); break;
        case 7: _t->currentChangedThroughResort((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 8: _t->searchTextChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 9: _t->searchRepeated(); break;
        case 10: _t->insertRow(); break;
        case 11: _t->deleteRows(); break;
        case 12: _t->editRow(); break;
        case 13: _t->togglePlayTracking(); break;
        case 14: _t->modelRemoved(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (ModelDataTableDialog::*)(sv_frame_t )>(_a, &ModelDataTableDialog::scrollToFrame, 0))
            return;
    }
}

const QMetaObject *sv::ModelDataTableDialog::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::ModelDataTableDialog::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv20ModelDataTableDialogE_t>.strings))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int sv::ModelDataTableDialog::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 15)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 15;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 15)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 15;
    }
    return _id;
}

// SIGNAL 0
void sv::ModelDataTableDialog::scrollToFrame(sv_frame_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}
QT_WARNING_POP
