/****************************************************************************
** Meta object code from reading C++ file 'Preferences.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svcore/base/Preferences.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'Preferences.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv11PreferencesE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::Preferences::qt_create_metaobjectdata<qt_meta_tag_ZN2sv11PreferencesE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::Preferences",
        "setProperty",
        "",
        "PropertyName",
        "setSpectrogramSmoothing",
        "SpectrogramSmoothing",
        "smoothing",
        "setTuningFrequency",
        "freq",
        "setPropertyBoxLayout",
        "PropertyBoxLayout",
        "layout",
        "setWindowType",
        "WindowType",
        "type",
        "setRunPluginsInProcess",
        "r",
        "setOmitTempsFromRecentFiles",
        "omit",
        "setTemporaryDirectoryRoot",
        "tempDirRoot",
        "setFixedSampleRate",
        "sv_samplerate_t",
        "setRecordMono",
        "setResampleOnLoad",
        "setUseGaplessMode",
        "setNormaliseAudio",
        "setFinerTimeStretch",
        "setBackgroundMode",
        "BackgroundMode",
        "mode",
        "setTimeToTextMode",
        "TimeToTextMode",
        "setTimeToTextModeUnsaved",
        "setShowHMS",
        "show",
        "setOctaveOfMiddleC",
        "oct",
        "setViewFontSize",
        "size",
        "setShowSplash"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'setProperty'
        QtMocHelpers::SlotData<void(const PropertyName &, int)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 }, { QMetaType::Int, 2 },
        }}),
        // Slot 'setSpectrogramSmoothing'
        QtMocHelpers::SlotData<void(enum SpectrogramSmoothing)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 6 },
        }}),
        // Slot 'setTuningFrequency'
        QtMocHelpers::SlotData<void(double)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 8 },
        }}),
        // Slot 'setPropertyBoxLayout'
        QtMocHelpers::SlotData<void(enum PropertyBoxLayout)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 },
        }}),
        // Slot 'setWindowType'
        QtMocHelpers::SlotData<void(WindowType)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 14 },
        }}),
        // Slot 'setRunPluginsInProcess'
        QtMocHelpers::SlotData<void(bool)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 16 },
        }}),
        // Slot 'setOmitTempsFromRecentFiles'
        QtMocHelpers::SlotData<void(bool)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 18 },
        }}),
        // Slot 'setTemporaryDirectoryRoot'
        QtMocHelpers::SlotData<void(QString)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 20 },
        }}),
        // Slot 'setFixedSampleRate'
        QtMocHelpers::SlotData<void(sv_samplerate_t)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 22, 2 },
        }}),
        // Slot 'setRecordMono'
        QtMocHelpers::SlotData<void(bool)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setResampleOnLoad'
        QtMocHelpers::SlotData<void(bool)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setUseGaplessMode'
        QtMocHelpers::SlotData<void(bool)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setNormaliseAudio'
        QtMocHelpers::SlotData<void(bool)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setFinerTimeStretch'
        QtMocHelpers::SlotData<void(bool)>(27, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setBackgroundMode'
        QtMocHelpers::SlotData<void(enum BackgroundMode)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 29, 30 },
        }}),
        // Slot 'setTimeToTextMode'
        QtMocHelpers::SlotData<void(enum TimeToTextMode)>(31, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 32, 30 },
        }}),
        // Slot 'setTimeToTextModeUnsaved'
        QtMocHelpers::SlotData<void(enum TimeToTextMode)>(33, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 32, 30 },
        }}),
        // Slot 'setShowHMS'
        QtMocHelpers::SlotData<void(bool)>(34, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 35 },
        }}),
        // Slot 'setOctaveOfMiddleC'
        QtMocHelpers::SlotData<void(int)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 37 },
        }}),
        // Slot 'setViewFontSize'
        QtMocHelpers::SlotData<void(int)>(38, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 39 },
        }}),
        // Slot 'setShowSplash'
        QtMocHelpers::SlotData<void(bool)>(40, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<Preferences, qt_meta_tag_ZN2sv11PreferencesE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::Preferences::staticMetaObject = { {
    QMetaObject::SuperData::link<PropertyContainer::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11PreferencesE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11PreferencesE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv11PreferencesE_t>.metaTypes,
    nullptr
} };

void sv::Preferences::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<Preferences *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->setProperty((*reinterpret_cast<std::add_pointer_t<PropertyName>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 1: _t->setSpectrogramSmoothing((*reinterpret_cast<std::add_pointer_t<enum SpectrogramSmoothing>>(_a[1]))); break;
        case 2: _t->setTuningFrequency((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 3: _t->setPropertyBoxLayout((*reinterpret_cast<std::add_pointer_t<enum PropertyBoxLayout>>(_a[1]))); break;
        case 4: _t->setWindowType((*reinterpret_cast<std::add_pointer_t<WindowType>>(_a[1]))); break;
        case 5: _t->setRunPluginsInProcess((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 6: _t->setOmitTempsFromRecentFiles((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 7: _t->setTemporaryDirectoryRoot((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 8: _t->setFixedSampleRate((*reinterpret_cast<std::add_pointer_t<sv_samplerate_t>>(_a[1]))); break;
        case 9: _t->setRecordMono((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 10: _t->setResampleOnLoad((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 11: _t->setUseGaplessMode((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 12: _t->setNormaliseAudio((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 13: _t->setFinerTimeStretch((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 14: _t->setBackgroundMode((*reinterpret_cast<std::add_pointer_t<enum BackgroundMode>>(_a[1]))); break;
        case 15: _t->setTimeToTextMode((*reinterpret_cast<std::add_pointer_t<enum TimeToTextMode>>(_a[1]))); break;
        case 16: _t->setTimeToTextModeUnsaved((*reinterpret_cast<std::add_pointer_t<enum TimeToTextMode>>(_a[1]))); break;
        case 17: _t->setShowHMS((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 18: _t->setOctaveOfMiddleC((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 19: _t->setViewFontSize((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 20: _t->setShowSplash((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObject *sv::Preferences::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::Preferences::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11PreferencesE_t>.strings))
        return static_cast<void*>(this);
    return PropertyContainer::qt_metacast(_clname);
}

int sv::Preferences::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = PropertyContainer::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 21)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 21;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 21)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 21;
    }
    return _id;
}
QT_WARNING_POP
