/****************************************************************************
** Meta object code from reading C++ file 'LayerTree.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/widgets/LayerTree.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'LayerTree.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv18ModelMetadataModelE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::ModelMetadataModel::qt_create_metaobjectdata<qt_meta_tag_ZN2sv18ModelMetadataModelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::ModelMetadataModel",
        "paneAdded",
        "",
        "paneDeleted",
        "propertyContainerAdded",
        "PropertyContainer*",
        "propertyContainerRemoved",
        "propertyContainerSelected",
        "propertyContainerPropertyChanged",
        "playParametersAudibilityChanged",
        "paneLayerModelChanged",
        "rebuildModelSet"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'paneAdded'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'paneDeleted'
        QtMocHelpers::SlotData<void()>(3, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'propertyContainerAdded'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(4, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 5, 2 },
        }}),
        // Slot 'propertyContainerRemoved'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(6, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 5, 2 },
        }}),
        // Slot 'propertyContainerSelected'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(7, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 5, 2 },
        }}),
        // Slot 'propertyContainerPropertyChanged'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(8, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 5, 2 },
        }}),
        // Slot 'playParametersAudibilityChanged'
        QtMocHelpers::SlotData<void(bool)>(9, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'paneLayerModelChanged'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'rebuildModelSet'
        QtMocHelpers::SlotData<void()>(11, 2, QMC::AccessProtected, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ModelMetadataModel, qt_meta_tag_ZN2sv18ModelMetadataModelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::ModelMetadataModel::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractItemModel::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv18ModelMetadataModelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv18ModelMetadataModelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv18ModelMetadataModelE_t>.metaTypes,
    nullptr
} };

void sv::ModelMetadataModel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ModelMetadataModel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->paneAdded(); break;
        case 1: _t->paneDeleted(); break;
        case 2: _t->propertyContainerAdded((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 3: _t->propertyContainerRemoved((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 4: _t->propertyContainerSelected((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 5: _t->propertyContainerPropertyChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 6: _t->playParametersAudibilityChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 7: _t->paneLayerModelChanged(); break;
        case 8: _t->rebuildModelSet(); break;
        default: ;
        }
    }
}

const QMetaObject *sv::ModelMetadataModel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::ModelMetadataModel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv18ModelMetadataModelE_t>.strings))
        return static_cast<void*>(this);
    return QAbstractItemModel::qt_metacast(_clname);
}

int sv::ModelMetadataModel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractItemModel::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 9)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 9;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 9)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 9;
    }
    return _id;
}
namespace {
struct qt_meta_tag_ZN2sv14LayerTreeModelE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::LayerTreeModel::qt_create_metaobjectdata<qt_meta_tag_ZN2sv14LayerTreeModelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::LayerTreeModel",
        "paneAdded",
        "",
        "paneAboutToBeDeleted",
        "Pane*",
        "propertyContainerAdded",
        "PropertyContainer*",
        "propertyContainerRemoved",
        "propertyContainerSelected",
        "propertyContainerPropertyChanged",
        "paneLayerModelChanged",
        "playParametersAudibilityChanged"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'paneAdded'
        QtMocHelpers::SlotData<void()>(1, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'paneAboutToBeDeleted'
        QtMocHelpers::SlotData<void(Pane *)>(3, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 4, 2 },
        }}),
        // Slot 'propertyContainerAdded'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(5, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'propertyContainerRemoved'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(7, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'propertyContainerSelected'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(8, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'propertyContainerPropertyChanged'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(9, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'paneLayerModelChanged'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playParametersAudibilityChanged'
        QtMocHelpers::SlotData<void(bool)>(11, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<LayerTreeModel, qt_meta_tag_ZN2sv14LayerTreeModelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::LayerTreeModel::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractItemModel::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14LayerTreeModelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14LayerTreeModelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv14LayerTreeModelE_t>.metaTypes,
    nullptr
} };

void sv::LayerTreeModel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<LayerTreeModel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->paneAdded(); break;
        case 1: _t->paneAboutToBeDeleted((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 2: _t->propertyContainerAdded((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 3: _t->propertyContainerRemoved((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 4: _t->propertyContainerSelected((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 5: _t->propertyContainerPropertyChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 6: _t->paneLayerModelChanged(); break;
        case 7: _t->playParametersAudibilityChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        default: ;
        }
    }
}

const QMetaObject *sv::LayerTreeModel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::LayerTreeModel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14LayerTreeModelE_t>.strings))
        return static_cast<void*>(this);
    return QAbstractItemModel::qt_metacast(_clname);
}

int sv::LayerTreeModel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractItemModel::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 8)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 8;
    }
    return _id;
}
QT_WARNING_POP
