/****************************************************************************
** Meta object code from reading C++ file 'PropertyContainer.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svcore/base/PropertyContainer.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'PropertyContainer.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv17PropertyContainerE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::PropertyContainer::qt_create_metaobjectdata<qt_meta_tag_ZN2sv17PropertyContainerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::PropertyContainer",
        "propertyChanged",
        "",
        "PropertyContainer::PropertyName",
        "setProperty",
        "PropertyName",
        "value",
        "setPropertyProtected",
        "getSetPropertyCommand",
        "Command*",
        "setPropertyFuzzy",
        "nameString",
        "valueString",
        "takeDiscretionaryPropertyMutex",
        "releaseDiscretionaryPropertyMutex"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'propertyChanged'
        QtMocHelpers::SignalData<void(PropertyContainer::PropertyName)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Slot 'setProperty'
        QtMocHelpers::SlotData<void(const PropertyName &, int)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 2 }, { QMetaType::Int, 6 },
        }}),
        // Slot 'setPropertyProtected'
        QtMocHelpers::SlotData<void(const PropertyName &, int)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 2 }, { QMetaType::Int, 6 },
        }}),
        // Slot 'getSetPropertyCommand'
        QtMocHelpers::SlotData<Command *(const PropertyName &, int)>(8, 2, QMC::AccessPublic, 0x80000000 | 9, {{
            { 0x80000000 | 5, 2 }, { QMetaType::Int, 6 },
        }}),
        // Slot 'setPropertyFuzzy'
        QtMocHelpers::SlotData<void(QString, QString)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 11 }, { QMetaType::QString, 12 },
        }}),
        // Slot 'getSetPropertyCommand'
        QtMocHelpers::SlotData<Command *(QString, QString)>(8, 2, QMC::AccessPublic, 0x80000000 | 9, {{
            { QMetaType::QString, 11 }, { QMetaType::QString, 12 },
        }}),
        // Slot 'takeDiscretionaryPropertyMutex'
        QtMocHelpers::SlotData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'releaseDiscretionaryPropertyMutex'
        QtMocHelpers::SlotData<void()>(14, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PropertyContainer, qt_meta_tag_ZN2sv17PropertyContainerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::PropertyContainer::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv17PropertyContainerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv17PropertyContainerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv17PropertyContainerE_t>.metaTypes,
    nullptr
} };

void sv::PropertyContainer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PropertyContainer *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->propertyChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer::PropertyName>>(_a[1]))); break;
        case 1: _t->setProperty((*reinterpret_cast<std::add_pointer_t<PropertyName>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 2: _t->setPropertyProtected((*reinterpret_cast<std::add_pointer_t<PropertyName>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 3: { Command* _r = _t->getSetPropertyCommand((*reinterpret_cast<std::add_pointer_t<PropertyName>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2])));
            if (_a[0]) *reinterpret_cast<Command**>(_a[0]) = std::move(_r); }  break;
        case 4: _t->setPropertyFuzzy((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 5: { Command* _r = _t->getSetPropertyCommand((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])));
            if (_a[0]) *reinterpret_cast<Command**>(_a[0]) = std::move(_r); }  break;
        case 6: _t->takeDiscretionaryPropertyMutex(); break;
        case 7: _t->releaseDiscretionaryPropertyMutex(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PropertyContainer::*)(PropertyContainer::PropertyName )>(_a, &PropertyContainer::propertyChanged, 0))
            return;
    }
}

const QMetaObject *sv::PropertyContainer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::PropertyContainer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv17PropertyContainerE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int sv::PropertyContainer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 8)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 8;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 8)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 8;
    }
    return _id;
}

// SIGNAL 0
void sv::PropertyContainer::propertyChanged(PropertyContainer::PropertyName _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}
QT_WARNING_POP
