/****************************************************************************
** Meta object code from reading C++ file 'Panner.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/widgets/Panner.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'Panner.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv6PannerE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::Panner::qt_create_metaobjectdata<qt_meta_tag_ZN2sv6PannerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::Panner",
        "rectExtentsChanged",
        "",
        "rectCentreMoved",
        "doubleClicked",
        "mouseEntered",
        "mouseLeft",
        "setRectExtents",
        "x0",
        "y0",
        "width",
        "height",
        "setRectWidth",
        "setRectHeight",
        "setRectCentreX",
        "x",
        "setRectCentreY",
        "y",
        "scroll",
        "up",
        "n",
        "resetToDefault"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'rectExtentsChanged'
        QtMocHelpers::SignalData<void(float, float, float, float)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 2 }, { QMetaType::Float, 2 }, { QMetaType::Float, 2 }, { QMetaType::Float, 2 },
        }}),
        // Signal 'rectCentreMoved'
        QtMocHelpers::SignalData<void(float, float)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 2 }, { QMetaType::Float, 2 },
        }}),
        // Signal 'doubleClicked'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'mouseEntered'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'mouseLeft'
        QtMocHelpers::SignalData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'setRectExtents'
        QtMocHelpers::SlotData<void(float, float, float, float)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 8 }, { QMetaType::Float, 9 }, { QMetaType::Float, 10 }, { QMetaType::Float, 11 },
        }}),
        // Slot 'setRectWidth'
        QtMocHelpers::SlotData<void(float)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 10 },
        }}),
        // Slot 'setRectHeight'
        QtMocHelpers::SlotData<void(float)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 11 },
        }}),
        // Slot 'setRectCentreX'
        QtMocHelpers::SlotData<void(float)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 15 },
        }}),
        // Slot 'setRectCentreY'
        QtMocHelpers::SlotData<void(float)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 17 },
        }}),
        // Slot 'scroll'
        QtMocHelpers::SlotData<void(bool)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 19 },
        }}),
        // Slot 'scroll'
        QtMocHelpers::SlotData<void(bool, int)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 19 }, { QMetaType::Int, 20 },
        }}),
        // Slot 'resetToDefault'
        QtMocHelpers::SlotData<void()>(21, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<Panner, qt_meta_tag_ZN2sv6PannerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::Panner::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv6PannerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv6PannerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv6PannerE_t>.metaTypes,
    nullptr
} };

void sv::Panner::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<Panner *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->rectExtentsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[4]))); break;
        case 1: _t->rectCentreMoved((*reinterpret_cast<std::add_pointer_t<float>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[2]))); break;
        case 2: _t->doubleClicked(); break;
        case 3: _t->mouseEntered(); break;
        case 4: _t->mouseLeft(); break;
        case 5: _t->setRectExtents((*reinterpret_cast<std::add_pointer_t<float>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[4]))); break;
        case 6: _t->setRectWidth((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 7: _t->setRectHeight((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 8: _t->setRectCentreX((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 9: _t->setRectCentreY((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 10: _t->scroll((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 11: _t->scroll((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        case 12: _t->resetToDefault(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (Panner::*)(float , float , float , float )>(_a, &Panner::rectExtentsChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (Panner::*)(float , float )>(_a, &Panner::rectCentreMoved, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (Panner::*)()>(_a, &Panner::doubleClicked, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (Panner::*)()>(_a, &Panner::mouseEntered, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (Panner::*)()>(_a, &Panner::mouseLeft, 4))
            return;
    }
}

const QMetaObject *sv::Panner::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::Panner::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv6PannerE_t>.strings))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int sv::Panner::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 13)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 13;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 13)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 13;
    }
    return _id;
}

// SIGNAL 0
void sv::Panner::rectExtentsChanged(float _t1, float _t2, float _t3, float _t4)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1, _t2, _t3, _t4);
}

// SIGNAL 1
void sv::Panner::rectCentreMoved(float _t1, float _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void sv::Panner::doubleClicked()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}

// SIGNAL 3
void sv::Panner::mouseEntered()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void sv::Panner::mouseLeft()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}
QT_WARNING_POP
