/****************************************************************************
** Meta object code from reading C++ file 'View.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/view/View.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'View.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv4ViewE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::View::qt_create_metaobjectdata<qt_meta_tag_ZN2sv4ViewE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::View",
        "propertyContainerAdded",
        "",
        "PropertyContainer*",
        "pc",
        "propertyContainerRemoved",
        "propertyContainerPropertyChanged",
        "propertyContainerPropertyRangeChanged",
        "propertyContainerNameChanged",
        "propertyContainerSelected",
        "propertyChanged",
        "PropertyContainer::PropertyName",
        "layerModelChanged",
        "cancelButtonPressed",
        "Layer*",
        "centreFrameChanged",
        "sv_frame_t",
        "frame",
        "globalScroll",
        "PlaybackFollowMode",
        "followMode",
        "zoomLevelChanged",
        "ZoomLevel",
        "level",
        "locked",
        "contextHelpChanged",
        "modelChanged",
        "ModelId",
        "modelChangedWithin",
        "startFrame",
        "endFrame",
        "modelCompletionChanged",
        "modelAlignmentCompletionChanged",
        "modelReplaced",
        "layerParametersChanged",
        "layerParameterRangesChanged",
        "layerMeasurementRectsChanged",
        "layerNameChanged",
        "globalCentreFrameChanged",
        "viewCentreFrameChanged",
        "View*",
        "viewManagerPlaybackFrameChanged",
        "viewZoomLevelChanged",
        "selectionChanged",
        "toolModeChanged",
        "overlayModeChanged",
        "zoomWheelsEnabledChanged",
        "cancelClicked",
        "progressCheckStalledTimerElapsed"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'propertyContainerAdded'
        QtMocHelpers::SignalData<void(PropertyContainer *)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'propertyContainerRemoved'
        QtMocHelpers::SignalData<void(PropertyContainer *)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'propertyContainerPropertyChanged'
        QtMocHelpers::SignalData<void(PropertyContainer *)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'propertyContainerPropertyRangeChanged'
        QtMocHelpers::SignalData<void(PropertyContainer *)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'propertyContainerNameChanged'
        QtMocHelpers::SignalData<void(PropertyContainer *)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'propertyContainerSelected'
        QtMocHelpers::SignalData<void(PropertyContainer *)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'propertyChanged'
        QtMocHelpers::SignalData<void(PropertyContainer::PropertyName)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 11, 2 },
        }}),
        // Signal 'layerModelChanged'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'cancelButtonPressed'
        QtMocHelpers::SignalData<void(Layer *)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 14, 2 },
        }}),
        // Signal 'centreFrameChanged'
        QtMocHelpers::SignalData<void(sv_frame_t, bool, PlaybackFollowMode)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 16, 17 }, { QMetaType::Bool, 18 }, { 0x80000000 | 19, 20 },
        }}),
        // Signal 'zoomLevelChanged'
        QtMocHelpers::SignalData<void(ZoomLevel, bool)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 22, 23 }, { QMetaType::Bool, 24 },
        }}),
        // Signal 'contextHelpChanged'
        QtMocHelpers::SignalData<void(const QString &)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'modelChanged'
        QtMocHelpers::SlotData<void(ModelId)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 2 },
        }}),
        // Slot 'modelChangedWithin'
        QtMocHelpers::SlotData<void(ModelId, sv_frame_t, sv_frame_t)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 2 }, { 0x80000000 | 16, 29 }, { 0x80000000 | 16, 30 },
        }}),
        // Slot 'modelCompletionChanged'
        QtMocHelpers::SlotData<void(ModelId)>(31, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 2 },
        }}),
        // Slot 'modelAlignmentCompletionChanged'
        QtMocHelpers::SlotData<void(ModelId)>(32, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 27, 2 },
        }}),
        // Slot 'modelReplaced'
        QtMocHelpers::SlotData<void()>(33, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'layerParametersChanged'
        QtMocHelpers::SlotData<void()>(34, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'layerParameterRangesChanged'
        QtMocHelpers::SlotData<void()>(35, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'layerMeasurementRectsChanged'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'layerNameChanged'
        QtMocHelpers::SlotData<void()>(37, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'globalCentreFrameChanged'
        QtMocHelpers::SlotData<void(sv_frame_t)>(38, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 16, 2 },
        }}),
        // Slot 'viewCentreFrameChanged'
        QtMocHelpers::SlotData<void(View *, sv_frame_t)>(39, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 40, 2 }, { 0x80000000 | 16, 2 },
        }}),
        // Slot 'viewManagerPlaybackFrameChanged'
        QtMocHelpers::SlotData<void(sv_frame_t)>(41, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 16, 2 },
        }}),
        // Slot 'viewZoomLevelChanged'
        QtMocHelpers::SlotData<void(View *, ZoomLevel, bool)>(42, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 40, 2 }, { 0x80000000 | 22, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Slot 'propertyContainerSelected'
        QtMocHelpers::SlotData<void(View *, PropertyContainer *)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 40, 2 }, { 0x80000000 | 3, 4 },
        }}),
        // Slot 'selectionChanged'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'toolModeChanged'
        QtMocHelpers::SlotData<void()>(44, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'overlayModeChanged'
        QtMocHelpers::SlotData<void()>(45, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'zoomWheelsEnabledChanged'
        QtMocHelpers::SlotData<void()>(46, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'cancelClicked'
        QtMocHelpers::SlotData<void()>(47, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'progressCheckStalledTimerElapsed'
        QtMocHelpers::SlotData<void()>(48, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<View, qt_meta_tag_ZN2sv4ViewE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::View::staticMetaObject = { {
    QMetaObject::SuperData::link<QFrame::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv4ViewE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv4ViewE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv4ViewE_t>.metaTypes,
    nullptr
} };

void sv::View::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<View *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->propertyContainerAdded((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 1: _t->propertyContainerRemoved((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 2: _t->propertyContainerPropertyChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 3: _t->propertyContainerPropertyRangeChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 4: _t->propertyContainerNameChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 5: _t->propertyContainerSelected((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 6: _t->propertyChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer::PropertyName>>(_a[1]))); break;
        case 7: _t->layerModelChanged(); break;
        case 8: _t->cancelButtonPressed((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1]))); break;
        case 9: _t->centreFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<PlaybackFollowMode>>(_a[3]))); break;
        case 10: _t->zoomLevelChanged((*reinterpret_cast<std::add_pointer_t<ZoomLevel>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 11: _t->contextHelpChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 12: _t->modelChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 13: _t->modelChangedWithin((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[3]))); break;
        case 14: _t->modelCompletionChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 15: _t->modelAlignmentCompletionChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 16: _t->modelReplaced(); break;
        case 17: _t->layerParametersChanged(); break;
        case 18: _t->layerParameterRangesChanged(); break;
        case 19: _t->layerMeasurementRectsChanged(); break;
        case 20: _t->layerNameChanged(); break;
        case 21: _t->globalCentreFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 22: _t->viewCentreFrameChanged((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2]))); break;
        case 23: _t->viewManagerPlaybackFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 24: _t->viewZoomLevelChanged((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<ZoomLevel>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 25: _t->propertyContainerSelected((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[2]))); break;
        case 26: _t->selectionChanged(); break;
        case 27: _t->toolModeChanged(); break;
        case 28: _t->overlayModeChanged(); break;
        case 29: _t->zoomWheelsEnabledChanged(); break;
        case 30: _t->cancelClicked(); break;
        case 31: _t->progressCheckStalledTimerElapsed(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 22:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< View* >(); break;
            }
            break;
        case 24:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< View* >(); break;
            }
            break;
        case 25:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< View* >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (View::*)(PropertyContainer * )>(_a, &View::propertyContainerAdded, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(PropertyContainer * )>(_a, &View::propertyContainerRemoved, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(PropertyContainer * )>(_a, &View::propertyContainerPropertyChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(PropertyContainer * )>(_a, &View::propertyContainerPropertyRangeChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(PropertyContainer * )>(_a, &View::propertyContainerNameChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(PropertyContainer * )>(_a, &View::propertyContainerSelected, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(PropertyContainer::PropertyName )>(_a, &View::propertyChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)()>(_a, &View::layerModelChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(Layer * )>(_a, &View::cancelButtonPressed, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(sv_frame_t , bool , PlaybackFollowMode )>(_a, &View::centreFrameChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(ZoomLevel , bool )>(_a, &View::zoomLevelChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (View::*)(const QString & )>(_a, &View::contextHelpChanged, 11))
            return;
    }
}

const QMetaObject *sv::View::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::View::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv4ViewE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "XmlExportable"))
        return static_cast< XmlExportable*>(this);
    if (!strcmp(_clname, "LayerGeometryProvider"))
        return static_cast< LayerGeometryProvider*>(this);
    return QFrame::qt_metacast(_clname);
}

int sv::View::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QFrame::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 32)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 32;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 32)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 32;
    }
    return _id;
}

// SIGNAL 0
void sv::View::propertyContainerAdded(PropertyContainer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::View::propertyContainerRemoved(PropertyContainer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void sv::View::propertyContainerPropertyChanged(PropertyContainer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::View::propertyContainerPropertyRangeChanged(PropertyContainer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void sv::View::propertyContainerNameChanged(PropertyContainer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void sv::View::propertyContainerSelected(PropertyContainer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void sv::View::propertyChanged(PropertyContainer::PropertyName _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void sv::View::layerModelChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void sv::View::cancelButtonPressed(Layer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void sv::View::centreFrameChanged(sv_frame_t _t1, bool _t2, PlaybackFollowMode _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1, _t2, _t3);
}

// SIGNAL 10
void sv::View::zoomLevelChanged(ZoomLevel _t1, bool _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1, _t2);
}

// SIGNAL 11
void sv::View::contextHelpChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1);
}
namespace {
struct qt_meta_tag_ZN2sv21ViewPropertyContainerE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::ViewPropertyContainer::qt_create_metaobjectdata<qt_meta_tag_ZN2sv21ViewPropertyContainerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::ViewPropertyContainer",
        "setProperty",
        "",
        "PropertyName",
        "n",
        "value"
    };

    QtMocHelpers::UintData qt_methods {
        // Slot 'setProperty'
        QtMocHelpers::SlotData<void(const PropertyName &, int)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { QMetaType::Int, 5 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ViewPropertyContainer, qt_meta_tag_ZN2sv21ViewPropertyContainerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::ViewPropertyContainer::staticMetaObject = { {
    QMetaObject::SuperData::link<PropertyContainer::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv21ViewPropertyContainerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv21ViewPropertyContainerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv21ViewPropertyContainerE_t>.metaTypes,
    nullptr
} };

void sv::ViewPropertyContainer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ViewPropertyContainer *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->setProperty((*reinterpret_cast<std::add_pointer_t<PropertyName>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<int>>(_a[2]))); break;
        default: ;
        }
    }
}

const QMetaObject *sv::ViewPropertyContainer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::ViewPropertyContainer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv21ViewPropertyContainerE_t>.strings))
        return static_cast<void*>(this);
    return PropertyContainer::qt_metacast(_clname);
}

int sv::ViewPropertyContainer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = PropertyContainer::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 1)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 1;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 1)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 1;
    }
    return _id;
}
QT_WARNING_POP
