/****************************************************************************
** Meta object code from reading C++ file 'PropertyStack.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/widgets/PropertyStack.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'PropertyStack.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv13PropertyStackE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::PropertyStack::qt_create_metaobjectdata<qt_meta_tag_ZN2sv13PropertyStackE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::PropertyStack",
        "viewSelected",
        "",
        "View*",
        "client",
        "propertyContainerSelected",
        "PropertyContainer*",
        "container",
        "propertyContainerContextMenuRequested",
        "QPoint",
        "pos",
        "contextHelpChanged",
        "propertyContainerAdded",
        "propertyContainerRemoved",
        "propertyContainerPropertyChanged",
        "propertyContainerPropertyRangeChanged",
        "propertyContainerNameChanged",
        "showLayer",
        "mouseEnteredTabBar",
        "mouseLeftTabBar",
        "activeTabClicked",
        "tabBarContextMenuRequested",
        "selectedContainerChanged"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'viewSelected'
        QtMocHelpers::SignalData<void(View *)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'propertyContainerSelected'
        QtMocHelpers::SignalData<void(View *, PropertyContainer *)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 6, 7 },
        }}),
        // Signal 'propertyContainerContextMenuRequested'
        QtMocHelpers::SignalData<void(View *, PropertyContainer *, QPoint)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 6, 7 }, { 0x80000000 | 9, 10 },
        }}),
        // Signal 'contextHelpChanged'
        QtMocHelpers::SignalData<void(const QString &)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'propertyContainerAdded'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'propertyContainerRemoved'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'propertyContainerPropertyChanged'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'propertyContainerPropertyRangeChanged'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'propertyContainerNameChanged'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Slot 'showLayer'
        QtMocHelpers::SlotData<void(bool)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'mouseEnteredTabBar'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'mouseLeftTabBar'
        QtMocHelpers::SlotData<void()>(19, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'activeTabClicked'
        QtMocHelpers::SlotData<void()>(20, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'tabBarContextMenuRequested'
        QtMocHelpers::SlotData<void(const QPoint &)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 2 },
        }}),
        // Slot 'selectedContainerChanged'
        QtMocHelpers::SlotData<void(int)>(22, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PropertyStack, qt_meta_tag_ZN2sv13PropertyStackE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::PropertyStack::staticMetaObject = { {
    QMetaObject::SuperData::link<QTabWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv13PropertyStackE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv13PropertyStackE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv13PropertyStackE_t>.metaTypes,
    nullptr
} };

void sv::PropertyStack::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PropertyStack *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->viewSelected((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1]))); break;
        case 1: _t->propertyContainerSelected((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[2]))); break;
        case 2: _t->propertyContainerContextMenuRequested((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[3]))); break;
        case 3: _t->contextHelpChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 4: _t->propertyContainerAdded((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 5: _t->propertyContainerRemoved((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 6: _t->propertyContainerPropertyChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 7: _t->propertyContainerPropertyRangeChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 8: _t->propertyContainerNameChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 9: _t->showLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 10: _t->mouseEnteredTabBar(); break;
        case 11: _t->mouseLeftTabBar(); break;
        case 12: _t->activeTabClicked(); break;
        case 13: _t->tabBarContextMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 14: _t->selectedContainerChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PropertyStack::*)(View * )>(_a, &PropertyStack::viewSelected, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PropertyStack::*)(View * , PropertyContainer * )>(_a, &PropertyStack::propertyContainerSelected, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PropertyStack::*)(View * , PropertyContainer * , QPoint )>(_a, &PropertyStack::propertyContainerContextMenuRequested, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PropertyStack::*)(const QString & )>(_a, &PropertyStack::contextHelpChanged, 3))
            return;
    }
}

const QMetaObject *sv::PropertyStack::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::PropertyStack::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv13PropertyStackE_t>.strings))
        return static_cast<void*>(this);
    return QTabWidget::qt_metacast(_clname);
}

int sv::PropertyStack::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QTabWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 15)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 15;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 15)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 15;
    }
    return _id;
}

// SIGNAL 0
void sv::PropertyStack::viewSelected(View * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::PropertyStack::propertyContainerSelected(View * _t1, PropertyContainer * _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void sv::PropertyStack::propertyContainerContextMenuRequested(View * _t1, PropertyContainer * _t2, QPoint _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2, _t3);
}

// SIGNAL 3
void sv::PropertyStack::contextHelpChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}
QT_WARNING_POP
