/****************************************************************************
** Meta object code from reading C++ file 'Layer.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/layer/Layer.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'Layer.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv5LayerE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::Layer::qt_create_metaobjectdata<qt_meta_tag_ZN2sv5LayerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::Layer",
        "modelChanged",
        "",
        "ModelId",
        "modelCompletionChanged",
        "modelAlignmentCompletionChanged",
        "modelChangedWithin",
        "sv_frame_t",
        "startFrame",
        "endFrame",
        "modelReplaced",
        "layerParametersChanged",
        "layerParameterRangesChanged",
        "layerMeasurementRectsChanged",
        "layerNameChanged",
        "verticalZoomChanged",
        "showLayer",
        "LayerGeometryProvider*",
        "show"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'modelChanged'
        QtMocHelpers::SignalData<void(ModelId)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'modelCompletionChanged'
        QtMocHelpers::SignalData<void(ModelId)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'modelAlignmentCompletionChanged'
        QtMocHelpers::SignalData<void(ModelId)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'modelChangedWithin'
        QtMocHelpers::SignalData<void(ModelId, sv_frame_t, sv_frame_t)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 }, { 0x80000000 | 7, 8 }, { 0x80000000 | 7, 9 },
        }}),
        // Signal 'modelReplaced'
        QtMocHelpers::SignalData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'layerParametersChanged'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'layerParameterRangesChanged'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'layerMeasurementRectsChanged'
        QtMocHelpers::SignalData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'layerNameChanged'
        QtMocHelpers::SignalData<void()>(14, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'verticalZoomChanged'
        QtMocHelpers::SignalData<void()>(15, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'showLayer'
        QtMocHelpers::SlotData<void(LayerGeometryProvider *, bool)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 17, 2 }, { QMetaType::Bool, 18 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<Layer, qt_meta_tag_ZN2sv5LayerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::Layer::staticMetaObject = { {
    QMetaObject::SuperData::link<PropertyContainer::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv5LayerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv5LayerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv5LayerE_t>.metaTypes,
    nullptr
} };

void sv::Layer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<Layer *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->modelChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 1: _t->modelCompletionChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 2: _t->modelAlignmentCompletionChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 3: _t->modelChangedWithin((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[3]))); break;
        case 4: _t->modelReplaced(); break;
        case 5: _t->layerParametersChanged(); break;
        case 6: _t->layerParameterRangesChanged(); break;
        case 7: _t->layerMeasurementRectsChanged(); break;
        case 8: _t->layerNameChanged(); break;
        case 9: _t->verticalZoomChanged(); break;
        case 10: _t->showLayer((*reinterpret_cast<std::add_pointer_t<LayerGeometryProvider*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (Layer::*)(ModelId )>(_a, &Layer::modelChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)(ModelId )>(_a, &Layer::modelCompletionChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)(ModelId )>(_a, &Layer::modelAlignmentCompletionChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)(ModelId , sv_frame_t , sv_frame_t )>(_a, &Layer::modelChangedWithin, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)()>(_a, &Layer::modelReplaced, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)()>(_a, &Layer::layerParametersChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)()>(_a, &Layer::layerParameterRangesChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)()>(_a, &Layer::layerMeasurementRectsChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)()>(_a, &Layer::layerNameChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (Layer::*)()>(_a, &Layer::verticalZoomChanged, 9))
            return;
    }
}

const QMetaObject *sv::Layer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::Layer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv5LayerE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "XmlExportable"))
        return static_cast< XmlExportable*>(this);
    return PropertyContainer::qt_metacast(_clname);
}

int sv::Layer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = PropertyContainer::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 11)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 11;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 11)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 11;
    }
    return _id;
}

// SIGNAL 0
void sv::Layer::modelChanged(ModelId _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::Layer::modelCompletionChanged(ModelId _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void sv::Layer::modelAlignmentCompletionChanged(ModelId _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::Layer::modelChangedWithin(ModelId _t1, sv_frame_t _t2, sv_frame_t _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2, _t3);
}

// SIGNAL 4
void sv::Layer::modelReplaced()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void sv::Layer::layerParametersChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void sv::Layer::layerParameterRangesChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void sv::Layer::layerMeasurementRectsChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void sv::Layer::layerNameChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void sv::Layer::verticalZoomChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}
QT_WARNING_POP
