/****************************************************************************
** Meta object code from reading C++ file 'ModelDataTableModel.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svcore/data/model/ModelDataTableModel.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'ModelDataTableModel.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv19ModelDataTableModelE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::ModelDataTableModel::qt_create_metaobjectdata<qt_meta_tag_ZN2sv19ModelDataTableModelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::ModelDataTableModel",
        "frameSelected",
        "",
        "addCommand",
        "Command*",
        "currentChanged",
        "QModelIndex",
        "modelRemoved",
        "modelChanged",
        "ModelId",
        "modelChangedWithin",
        "sv_frame_t"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'frameSelected'
        QtMocHelpers::SignalData<void(int)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Signal 'addCommand'
        QtMocHelpers::SignalData<void(Command *)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 2 },
        }}),
        // Signal 'currentChanged'
        QtMocHelpers::SignalData<void(const QModelIndex &)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 2 },
        }}),
        // Signal 'modelRemoved'
        QtMocHelpers::SignalData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'modelChanged'
        QtMocHelpers::SlotData<void(ModelId)>(8, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 9, 2 },
        }}),
        // Slot 'modelChangedWithin'
        QtMocHelpers::SlotData<void(ModelId, sv_frame_t, sv_frame_t)>(10, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 9, 2 }, { 0x80000000 | 11, 2 }, { 0x80000000 | 11, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ModelDataTableModel, qt_meta_tag_ZN2sv19ModelDataTableModelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::ModelDataTableModel::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractItemModel::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv19ModelDataTableModelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv19ModelDataTableModelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv19ModelDataTableModelE_t>.metaTypes,
    nullptr
} };

void sv::ModelDataTableModel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ModelDataTableModel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->frameSelected((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 1: _t->addCommand((*reinterpret_cast<std::add_pointer_t<Command*>>(_a[1]))); break;
        case 2: _t->currentChanged((*reinterpret_cast<std::add_pointer_t<QModelIndex>>(_a[1]))); break;
        case 3: _t->modelRemoved(); break;
        case 4: _t->modelChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 5: _t->modelChangedWithin((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[3]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (ModelDataTableModel::*)(int )>(_a, &ModelDataTableModel::frameSelected, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (ModelDataTableModel::*)(Command * )>(_a, &ModelDataTableModel::addCommand, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (ModelDataTableModel::*)(const QModelIndex & )>(_a, &ModelDataTableModel::currentChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (ModelDataTableModel::*)()>(_a, &ModelDataTableModel::modelRemoved, 3))
            return;
    }
}

const QMetaObject *sv::ModelDataTableModel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::ModelDataTableModel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv19ModelDataTableModelE_t>.strings))
        return static_cast<void*>(this);
    return QAbstractItemModel::qt_metacast(_clname);
}

int sv::ModelDataTableModel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractItemModel::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 6)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 6;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 6)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 6;
    }
    return _id;
}

// SIGNAL 0
void sv::ModelDataTableModel::frameSelected(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::ModelDataTableModel::addCommand(Command * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void sv::ModelDataTableModel::currentChanged(const QModelIndex & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::ModelDataTableModel::modelRemoved()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}
QT_WARNING_POP
