/****************************************************************************
** Meta object code from reading C++ file 'PreferencesDialog.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/main/PreferencesDialog.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'PreferencesDialog.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN17PreferencesDialogE_t {};
} // unnamed namespace

template <> constexpr inline auto PreferencesDialog::qt_create_metaobjectdata<qt_meta_tag_ZN17PreferencesDialogE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "PreferencesDialog",
        "audioDeviceChanged",
        "",
        "coloursChanged",
        "applicationClosing",
        "quickly",
        "windowTypeChanged",
        "sv::WindowType",
        "type",
        "spectrogramSmoothingChanged",
        "state",
        "spectrogramGColourChanged",
        "spectrogramMColourChanged",
        "colour3DColourChanged",
        "overviewColourChanged",
        "propertyLayoutChanged",
        "layout",
        "tuningFrequencyChanged",
        "freq",
        "audioImplementationChanged",
        "impl",
        "audioPlaybackDeviceChanged",
        "device",
        "audioRecordDeviceChanged",
        "resampleOnLoadChanged",
        "finerTimeStretchChanged",
        "gaplessModeChanged",
        "vampProcessSeparationChanged",
        "tempDirRootChanged",
        "root",
        "backgroundModeChanged",
        "mode",
        "timeToTextModeChanged",
        "showHMSChanged",
        "octaveSystemChanged",
        "system",
        "viewFontSizeChanged",
        "sz",
        "showSplashChanged",
        "defaultTemplateChanged",
        "localeChanged",
        "networkPermissionChanged",
        "retinaChanged",
        "pluginPathsChanged",
        "tempDirButtonClicked",
        "okClicked",
        "applyClicked",
        "cancelClicked"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'audioDeviceChanged'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'coloursChanged'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'applicationClosing'
        QtMocHelpers::SlotData<void(bool)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 5 },
        }}),
        // Slot 'windowTypeChanged'
        QtMocHelpers::SlotData<void(sv::WindowType)>(6, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 7, 8 },
        }}),
        // Slot 'spectrogramSmoothingChanged'
        QtMocHelpers::SlotData<void(int)>(9, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'spectrogramGColourChanged'
        QtMocHelpers::SlotData<void(int)>(11, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'spectrogramMColourChanged'
        QtMocHelpers::SlotData<void(int)>(12, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'colour3DColourChanged'
        QtMocHelpers::SlotData<void(int)>(13, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'overviewColourChanged'
        QtMocHelpers::SlotData<void(int)>(14, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'propertyLayoutChanged'
        QtMocHelpers::SlotData<void(int)>(15, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 16 },
        }}),
        // Slot 'tuningFrequencyChanged'
        QtMocHelpers::SlotData<void(double)>(17, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Double, 18 },
        }}),
        // Slot 'audioImplementationChanged'
        QtMocHelpers::SlotData<void(int)>(19, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 20 },
        }}),
        // Slot 'audioPlaybackDeviceChanged'
        QtMocHelpers::SlotData<void(int)>(21, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 22 },
        }}),
        // Slot 'audioRecordDeviceChanged'
        QtMocHelpers::SlotData<void(int)>(23, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 22 },
        }}),
        // Slot 'resampleOnLoadChanged'
        QtMocHelpers::SlotData<void(int)>(24, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'finerTimeStretchChanged'
        QtMocHelpers::SlotData<void(int)>(25, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'gaplessModeChanged'
        QtMocHelpers::SlotData<void(int)>(26, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'vampProcessSeparationChanged'
        QtMocHelpers::SlotData<void(int)>(27, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'tempDirRootChanged'
        QtMocHelpers::SlotData<void(QString)>(28, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 29 },
        }}),
        // Slot 'backgroundModeChanged'
        QtMocHelpers::SlotData<void(int)>(30, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 31 },
        }}),
        // Slot 'timeToTextModeChanged'
        QtMocHelpers::SlotData<void(int)>(32, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 31 },
        }}),
        // Slot 'showHMSChanged'
        QtMocHelpers::SlotData<void(int)>(33, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'octaveSystemChanged'
        QtMocHelpers::SlotData<void(int)>(34, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 35 },
        }}),
        // Slot 'viewFontSizeChanged'
        QtMocHelpers::SlotData<void(int)>(36, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 37 },
        }}),
        // Slot 'showSplashChanged'
        QtMocHelpers::SlotData<void(int)>(38, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'defaultTemplateChanged'
        QtMocHelpers::SlotData<void(int)>(39, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'localeChanged'
        QtMocHelpers::SlotData<void(int)>(40, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'networkPermissionChanged'
        QtMocHelpers::SlotData<void(int)>(41, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'retinaChanged'
        QtMocHelpers::SlotData<void(int)>(42, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'pluginPathsChanged'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'tempDirButtonClicked'
        QtMocHelpers::SlotData<void()>(44, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'okClicked'
        QtMocHelpers::SlotData<void()>(45, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'applyClicked'
        QtMocHelpers::SlotData<void()>(46, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'cancelClicked'
        QtMocHelpers::SlotData<void()>(47, 2, QMC::AccessProtected, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PreferencesDialog, qt_meta_tag_ZN17PreferencesDialogE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject PreferencesDialog::staticMetaObject = { {
    QMetaObject::SuperData::link<QDialog::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17PreferencesDialogE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17PreferencesDialogE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN17PreferencesDialogE_t>.metaTypes,
    nullptr
} };

void PreferencesDialog::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PreferencesDialog *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->audioDeviceChanged(); break;
        case 1: _t->coloursChanged(); break;
        case 2: _t->applicationClosing((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 3: _t->windowTypeChanged((*reinterpret_cast<std::add_pointer_t<sv::WindowType>>(_a[1]))); break;
        case 4: _t->spectrogramSmoothingChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 5: _t->spectrogramGColourChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 6: _t->spectrogramMColourChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 7: _t->colour3DColourChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 8: _t->overviewColourChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 9: _t->propertyLayoutChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 10: _t->tuningFrequencyChanged((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 11: _t->audioImplementationChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 12: _t->audioPlaybackDeviceChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 13: _t->audioRecordDeviceChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 14: _t->resampleOnLoadChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 15: _t->finerTimeStretchChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 16: _t->gaplessModeChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 17: _t->vampProcessSeparationChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 18: _t->tempDirRootChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 19: _t->backgroundModeChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 20: _t->timeToTextModeChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 21: _t->showHMSChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 22: _t->octaveSystemChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 23: _t->viewFontSizeChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 24: _t->showSplashChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 25: _t->defaultTemplateChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 26: _t->localeChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 27: _t->networkPermissionChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 28: _t->retinaChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 29: _t->pluginPathsChanged(); break;
        case 30: _t->tempDirButtonClicked(); break;
        case 31: _t->okClicked(); break;
        case 32: _t->applyClicked(); break;
        case 33: _t->cancelClicked(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PreferencesDialog::*)()>(_a, &PreferencesDialog::audioDeviceChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PreferencesDialog::*)()>(_a, &PreferencesDialog::coloursChanged, 1))
            return;
    }
}

const QMetaObject *PreferencesDialog::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *PreferencesDialog::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN17PreferencesDialogE_t>.strings))
        return static_cast<void*>(this);
    return QDialog::qt_metacast(_clname);
}

int PreferencesDialog::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDialog::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 34)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 34;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 34)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 34;
    }
    return _id;
}

// SIGNAL 0
void PreferencesDialog::audioDeviceChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void PreferencesDialog::coloursChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}
QT_WARNING_POP
