/****************************************************************************
** Meta object code from reading C++ file 'ViewManager.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/view/ViewManager.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'ViewManager.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv11ViewManagerE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::ViewManager::qt_create_metaobjectdata<qt_meta_tag_ZN2sv11ViewManagerE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::ViewManager",
        "globalCentreFrameChanged",
        "",
        "sv_frame_t",
        "frame",
        "viewCentreFrameChanged",
        "View*",
        "v",
        "viewZoomLevelChanged",
        "ZoomLevel",
        "zoom",
        "locked",
        "playbackFrameChanged",
        "monitoringLevelsChanged",
        "left",
        "right",
        "selectionChanged",
        "selectionChangedByUser",
        "inProgressSelectionChanged",
        "toolModeChanged",
        "playLoopModeChanged",
        "playSelectionModeChanged",
        "playSoloModeChanged",
        "alignModeChanged",
        "overlayModeChanged",
        "showCentreLineChanged",
        "zoomWheelsEnabledChanged",
        "opportunisticEditingEnabledChanged",
        "activity",
        "PlaybackFollowMode",
        "setGlobalCentreFrame",
        "setPlaybackFrame",
        "playStatusChanged",
        "playing",
        "recordStatusChanged",
        "recording",
        "checkPlayStatus",
        "seek"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'globalCentreFrameChanged'
        QtMocHelpers::SignalData<void(sv_frame_t)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'viewCentreFrameChanged'
        QtMocHelpers::SignalData<void(View *, sv_frame_t)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 7 }, { 0x80000000 | 3, 4 },
        }}),
        // Signal 'viewZoomLevelChanged'
        QtMocHelpers::SignalData<void(View *, ZoomLevel, bool)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 6, 7 }, { 0x80000000 | 9, 10 }, { QMetaType::Bool, 11 },
        }}),
        // Signal 'playbackFrameChanged'
        QtMocHelpers::SignalData<void(sv_frame_t)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'monitoringLevelsChanged'
        QtMocHelpers::SignalData<void(float, float)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 14 }, { QMetaType::Float, 15 },
        }}),
        // Signal 'selectionChanged'
        QtMocHelpers::SignalData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'selectionChangedByUser'
        QtMocHelpers::SignalData<void()>(17, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'inProgressSelectionChanged'
        QtMocHelpers::SignalData<void()>(18, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'toolModeChanged'
        QtMocHelpers::SignalData<void()>(19, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playLoopModeChanged'
        QtMocHelpers::SignalData<void()>(20, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playLoopModeChanged'
        QtMocHelpers::SignalData<void(bool)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'playSelectionModeChanged'
        QtMocHelpers::SignalData<void()>(21, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playSelectionModeChanged'
        QtMocHelpers::SignalData<void(bool)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'playSoloModeChanged'
        QtMocHelpers::SignalData<void()>(22, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'playSoloModeChanged'
        QtMocHelpers::SignalData<void(bool)>(22, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'alignModeChanged'
        QtMocHelpers::SignalData<void()>(23, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'alignModeChanged'
        QtMocHelpers::SignalData<void(bool)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'overlayModeChanged'
        QtMocHelpers::SignalData<void()>(24, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'showCentreLineChanged'
        QtMocHelpers::SignalData<void()>(25, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'zoomWheelsEnabledChanged'
        QtMocHelpers::SignalData<void()>(26, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'opportunisticEditingEnabledChanged'
        QtMocHelpers::SignalData<void()>(27, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'activity'
        QtMocHelpers::SignalData<void(QString)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'viewCentreFrameChanged'
        QtMocHelpers::SlotData<void(sv_frame_t, bool, PlaybackFollowMode)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 }, { QMetaType::Bool, 2 }, { 0x80000000 | 29, 2 },
        }}),
        // Slot 'viewZoomLevelChanged'
        QtMocHelpers::SlotData<void(ZoomLevel, bool)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Slot 'setGlobalCentreFrame'
        QtMocHelpers::SlotData<void(sv_frame_t)>(30, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Slot 'setPlaybackFrame'
        QtMocHelpers::SlotData<void(sv_frame_t)>(31, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Slot 'playStatusChanged'
        QtMocHelpers::SlotData<void(bool)>(32, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 33 },
        }}),
        // Slot 'recordStatusChanged'
        QtMocHelpers::SlotData<void(bool)>(34, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 35 },
        }}),
        // Slot 'checkPlayStatus'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'seek'
        QtMocHelpers::SlotData<void(sv_frame_t)>(37, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<ViewManager, qt_meta_tag_ZN2sv11ViewManagerE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::ViewManager::staticMetaObject = { {
    QMetaObject::SuperData::link<ViewManagerBase::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11ViewManagerE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11ViewManagerE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv11ViewManagerE_t>.metaTypes,
    nullptr
} };

void sv::ViewManager::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<ViewManager *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->globalCentreFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 1: _t->viewCentreFrameChanged((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2]))); break;
        case 2: _t->viewZoomLevelChanged((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<ZoomLevel>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 3: _t->playbackFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 4: _t->monitoringLevelsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[2]))); break;
        case 5: _t->selectionChanged(); break;
        case 6: _t->selectionChangedByUser(); break;
        case 7: _t->inProgressSelectionChanged(); break;
        case 8: _t->toolModeChanged(); break;
        case 9: _t->playLoopModeChanged(); break;
        case 10: _t->playLoopModeChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 11: _t->playSelectionModeChanged(); break;
        case 12: _t->playSelectionModeChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 13: _t->playSoloModeChanged(); break;
        case 14: _t->playSoloModeChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 15: _t->alignModeChanged(); break;
        case 16: _t->alignModeChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 17: _t->overlayModeChanged(); break;
        case 18: _t->showCentreLineChanged(); break;
        case 19: _t->zoomWheelsEnabledChanged(); break;
        case 20: _t->opportunisticEditingEnabledChanged(); break;
        case 21: _t->activity((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 22: _t->viewCentreFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<PlaybackFollowMode>>(_a[3]))); break;
        case 23: _t->viewZoomLevelChanged((*reinterpret_cast<std::add_pointer_t<ZoomLevel>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 24: _t->setGlobalCentreFrame((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 25: _t->setPlaybackFrame((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 26: _t->playStatusChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 27: _t->recordStatusChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 28: _t->checkPlayStatus(); break;
        case 29: _t->seek((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(sv_frame_t )>(_a, &ViewManager::globalCentreFrameChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(View * , sv_frame_t )>(_a, &ViewManager::viewCentreFrameChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(View * , ZoomLevel , bool )>(_a, &ViewManager::viewZoomLevelChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(sv_frame_t )>(_a, &ViewManager::playbackFrameChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(float , float )>(_a, &ViewManager::monitoringLevelsChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::selectionChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::selectionChangedByUser, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::inProgressSelectionChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::toolModeChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::playLoopModeChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(bool )>(_a, &ViewManager::playLoopModeChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::playSelectionModeChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(bool )>(_a, &ViewManager::playSelectionModeChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::playSoloModeChanged, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(bool )>(_a, &ViewManager::playSoloModeChanged, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::alignModeChanged, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(bool )>(_a, &ViewManager::alignModeChanged, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::overlayModeChanged, 17))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::showCentreLineChanged, 18))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::zoomWheelsEnabledChanged, 19))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)()>(_a, &ViewManager::opportunisticEditingEnabledChanged, 20))
            return;
        if (QtMocHelpers::indexOfMethod<void (ViewManager::*)(QString )>(_a, &ViewManager::activity, 21))
            return;
    }
}

const QMetaObject *sv::ViewManager::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::ViewManager::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11ViewManagerE_t>.strings))
        return static_cast<void*>(this);
    return ViewManagerBase::qt_metacast(_clname);
}

int sv::ViewManager::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = ViewManagerBase::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 30)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 30;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 30)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 30;
    }
    return _id;
}

// SIGNAL 0
void sv::ViewManager::globalCentreFrameChanged(sv_frame_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::ViewManager::viewCentreFrameChanged(View * _t1, sv_frame_t _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void sv::ViewManager::viewZoomLevelChanged(View * _t1, ZoomLevel _t2, bool _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2, _t3);
}

// SIGNAL 3
void sv::ViewManager::playbackFrameChanged(sv_frame_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void sv::ViewManager::monitoringLevelsChanged(float _t1, float _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1, _t2);
}

// SIGNAL 5
void sv::ViewManager::selectionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 5, nullptr);
}

// SIGNAL 6
void sv::ViewManager::selectionChangedByUser()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void sv::ViewManager::inProgressSelectionChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void sv::ViewManager::toolModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void sv::ViewManager::playLoopModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void sv::ViewManager::playLoopModeChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void sv::ViewManager::playSelectionModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}

// SIGNAL 12
void sv::ViewManager::playSelectionModeChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1);
}

// SIGNAL 13
void sv::ViewManager::playSoloModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}

// SIGNAL 14
void sv::ViewManager::playSoloModeChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1);
}

// SIGNAL 15
void sv::ViewManager::alignModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 15, nullptr);
}

// SIGNAL 16
void sv::ViewManager::alignModeChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 16, nullptr, _t1);
}

// SIGNAL 17
void sv::ViewManager::overlayModeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 17, nullptr);
}

// SIGNAL 18
void sv::ViewManager::showCentreLineChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 18, nullptr);
}

// SIGNAL 19
void sv::ViewManager::zoomWheelsEnabledChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 19, nullptr);
}

// SIGNAL 20
void sv::ViewManager::opportunisticEditingEnabledChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 20, nullptr);
}

// SIGNAL 21
void sv::ViewManager::activity(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 21, nullptr, _t1);
}
QT_WARNING_POP
