/* -*- c-basic-offset: 4 indent-tabs-mode: nil -*-  vi:set ts=8 sts=4 sw=4: */

/*
    Sonic Visualiser
    An audio file viewer and annotation editor.
    Centre for Digital Music, Queen Mary, University of London.
    This file copyright 2006 Chris Cannam and QMUL.
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.  See the file
    COPYING included with this distribution for more information.
*/

#include "PropertyContainer.h"
#include "RangeMapper.h"
#include "UnitDatabase.h"

#include <QMutexLocker>

#include <iostream>

using std::cerr;
using std::endl;

namespace sv {

PropertyContainer::PropertyList
PropertyContainer::getProperties() const
{
    return PropertyList();
}

PropertyContainer::PropertyType
PropertyContainer::getPropertyType(const PropertyName &) const
{
    return InvalidProperty;
}

QString
PropertyContainer::getPropertyIconName(const PropertyName &) const
{
    return QString();
}

QString
PropertyContainer::getPropertyGroupName(const PropertyName &) const
{
    return QString();
}

int
PropertyContainer::getPropertyRangeAndValue(const PropertyName &,
                                            int *min, int *max, int *deflt) const
{
    if (min) *min = 0;
    if (max) *max = 0;
    if (deflt) *deflt = 0;
    return 0;
}

QString
PropertyContainer::getPropertyValueLabel(const PropertyName &, int) const
{
    return QString();
}

QString
PropertyContainer::getPropertyValueIconName(const PropertyName &, int) const
{
    return QString();
}

RangeMapper *
PropertyContainer::getNewPropertyRangeMapper(const PropertyName &) const
{
    return nullptr;
}

void
PropertyContainer::setProperty(const PropertyName &name, int) 
{
    cerr << "WARNING: PropertyContainer[" << getPropertyContainerName() << "]::setProperty(" << name << "): no implementation in subclass!" << endl;
}

void
PropertyContainer::setPropertyProtected(const PropertyName &name, int value) 
{
    QMutexLocker locker(&m_discretionaryPropertyMutex);
    setProperty(name, value);
}

Command *
PropertyContainer::getSetPropertyCommand(const PropertyName &name, int value)
{
    int currentValue = getPropertyRangeAndValue(name, nullptr, nullptr, nullptr);
    if (value == currentValue) return nullptr;
    return new SetPropertyCommand(this, name, value);
}
 
void
PropertyContainer::setPropertyFuzzy(QString nameString, QString valueString)
{
    PropertyName name;
    int value;
    if (!convertPropertyStrings(nameString, valueString, name, value)) {
        cerr << "WARNING: PropertyContainer::setProperty(\""
                  << nameString << "\", \""
                  << valueString
                  << "\"): Name and value conversion failed" << endl;
        return;
    }
    setProperty(name, value);
}
 
Command *
PropertyContainer::getSetPropertyCommand(QString nameString, QString valueString)
{
    PropertyName name;
    int value;
    if (!convertPropertyStrings(nameString, valueString, name, value)) {
        cerr << "WARNING: PropertyContainer::getSetPropertyCommand(\""
                  << nameString << "\", \""
                  << valueString
                  << "\"): Name and value conversion failed" << endl;
        return nullptr;
    }
    return getSetPropertyCommand(name, value);
}

bool
PropertyContainer::convertPropertyStrings(QString nameString, QString valueString,
                                          PropertyName &name, int &value)
{
    PropertyList pl = getProperties();

    QString adjusted = nameString.trimmed();
    adjusted.replace('_', ' ');
    adjusted.replace('-', ' ');
    
    name = "";

    for (PropertyList::iterator pli = pl.begin(); pli != pl.end(); ++pli) {

        QString label = getPropertyLabel(*pli);

        if (label != "" && (nameString == label || adjusted == label)) {
            name = *pli;
            break;
        } else if (nameString == *pli) {
            name = *pli;
            break;
        }
    }

    if (name == "") {
        cerr << "PropertyContainer::convertPropertyStrings: Unable to match name string \"" << nameString << "\"" << endl;
        return false;
    }

    value = 0;
    bool success = false;
    
    bool isDouble = false;
    double dval = valueString.toDouble(&isDouble);

    switch (getPropertyType(name)) {

    case ToggleProperty:
        if (valueString == tr("yes") || 
            valueString == tr("on") ||
            valueString == tr("true")) {
            value = 1; success = true;
        } else if (valueString == tr("no") ||
                   valueString == tr("off") ||
                   valueString == tr("false")) {
            value = 0; success = true;
        }
        break;

    case RangeProperty:
        if (isDouble) {
            RangeMapper *mapper = getNewPropertyRangeMapper(name);
            if (mapper) {
                value = mapper->getPositionForValue(dval);
                delete mapper;
                success = true;
            }
        }
        break;

    case ValueProperty:
    case ColourProperty:
    case ColourMapProperty:
    {
        int min, max;
        getPropertyRangeAndValue(name, &min, &max, nullptr);
        for (int i = min; i <= max; ++i) {
            if (valueString == getPropertyValueLabel(name, i)) {
                value = i;
                success = true;
                break;
            }
        }
        break;
    }
        
    case UnitsProperty:
        value = UnitDatabase::getInstance()->getUnitId(valueString, false);
        if (value >= 0) success = true;
        else value = 0;
        break;

    case InvalidProperty:
        SVDEBUG << "PropertyContainer::convertPropertyStrings: Invalid property name \"" << name << "\"" << endl;
        return false;
    }

    if (success) return true;

    int min, max;
    getPropertyRangeAndValue(name, &min, &max, nullptr);
    
    bool ok = false;
    int i = valueString.toInt(&ok);
    if (!ok) {
        cerr << "PropertyContainer::convertPropertyStrings: Unable to parse value string \"" << valueString << "\"" << endl;
        return false;
    } else if (i < min || i > max) {
        SVDEBUG << "PropertyContainer::convertPropertyStrings: Property value \"" << i << "\" outside valid range " << min << " to " << max << endl;
        return false;
    }

    value = i;
    return true;
}

void
PropertyContainer::takeDiscretionaryPropertyMutex()
{
    m_discretionaryPropertyMutex.lock();
}

void
PropertyContainer::releaseDiscretionaryPropertyMutex()
{
    m_discretionaryPropertyMutex.unlock();
}

PropertyContainer::SetPropertyCommand::SetPropertyCommand(PropertyContainer *pc,
                                                          const PropertyName &pn,
                                                          int value) :
    m_pc(pc),
    m_pn(pn),
    m_value(value),
    m_oldValue(0)
{
}

void
PropertyContainer::SetPropertyCommand::execute()
{
    m_oldValue = m_pc->getPropertyRangeAndValue(m_pn, nullptr, nullptr, nullptr);
    m_pc->setPropertyProtected(m_pn, m_value);
}

void
PropertyContainer::SetPropertyCommand::unexecute() 
{
    m_pc->setPropertyProtected(m_pn, m_oldValue);
}

QString
PropertyContainer::SetPropertyCommand::getName() const
{
    return tr("Set %1 Property").arg(m_pn);
}

} // end namespace sv

