/****************************************************************************
** Meta object code from reading C++ file 'CommandHistory.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/widgets/CommandHistory.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'CommandHistory.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv14CommandHistoryE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::CommandHistory::qt_create_metaobjectdata<qt_meta_tag_ZN2sv14CommandHistoryE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::CommandHistory",
        "commandExecuted",
        "",
        "Command*",
        "commandUnexecuted",
        "documentRestored",
        "activity",
        "documentSaved",
        "addExecutedCommand",
        "addCommandAndExecute",
        "undo",
        "redo",
        "undoActivated",
        "QAction*",
        "redoActivated",
        "bundleTimerTimeout"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'commandExecuted'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'commandExecuted'
        QtMocHelpers::SignalData<void(Command *)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'commandUnexecuted'
        QtMocHelpers::SignalData<void(Command *)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'documentRestored'
        QtMocHelpers::SignalData<void()>(5, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'activity'
        QtMocHelpers::SignalData<void(QString)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'documentSaved'
        QtMocHelpers::SlotData<void()>(7, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'addExecutedCommand'
        QtMocHelpers::SlotData<void(Command *)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Slot 'addCommandAndExecute'
        QtMocHelpers::SlotData<void(Command *)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Slot 'undo'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'redo'
        QtMocHelpers::SlotData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'undoActivated'
        QtMocHelpers::SlotData<void(QAction *)>(12, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 13, 2 },
        }}),
        // Slot 'redoActivated'
        QtMocHelpers::SlotData<void(QAction *)>(14, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 13, 2 },
        }}),
        // Slot 'bundleTimerTimeout'
        QtMocHelpers::SlotData<void()>(15, 2, QMC::AccessProtected, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<CommandHistory, qt_meta_tag_ZN2sv14CommandHistoryE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::CommandHistory::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14CommandHistoryE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14CommandHistoryE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv14CommandHistoryE_t>.metaTypes,
    nullptr
} };

void sv::CommandHistory::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<CommandHistory *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->commandExecuted(); break;
        case 1: _t->commandExecuted((*reinterpret_cast<std::add_pointer_t<Command*>>(_a[1]))); break;
        case 2: _t->commandUnexecuted((*reinterpret_cast<std::add_pointer_t<Command*>>(_a[1]))); break;
        case 3: _t->documentRestored(); break;
        case 4: _t->activity((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 5: _t->documentSaved(); break;
        case 6: _t->addExecutedCommand((*reinterpret_cast<std::add_pointer_t<Command*>>(_a[1]))); break;
        case 7: _t->addCommandAndExecute((*reinterpret_cast<std::add_pointer_t<Command*>>(_a[1]))); break;
        case 8: _t->undo(); break;
        case 9: _t->redo(); break;
        case 10: _t->undoActivated((*reinterpret_cast<std::add_pointer_t<QAction*>>(_a[1]))); break;
        case 11: _t->redoActivated((*reinterpret_cast<std::add_pointer_t<QAction*>>(_a[1]))); break;
        case 12: _t->bundleTimerTimeout(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (CommandHistory::*)()>(_a, &CommandHistory::commandExecuted, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (CommandHistory::*)(Command * )>(_a, &CommandHistory::commandExecuted, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (CommandHistory::*)(Command * )>(_a, &CommandHistory::commandUnexecuted, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (CommandHistory::*)()>(_a, &CommandHistory::documentRestored, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (CommandHistory::*)(QString )>(_a, &CommandHistory::activity, 4))
            return;
    }
}

const QMetaObject *sv::CommandHistory::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::CommandHistory::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14CommandHistoryE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int sv::CommandHistory::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 13)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 13;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 13)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 13;
    }
    return _id;
}

// SIGNAL 0
void sv::CommandHistory::commandExecuted()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void sv::CommandHistory::commandExecuted(Command * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void sv::CommandHistory::commandUnexecuted(Command * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::CommandHistory::documentRestored()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void sv::CommandHistory::activity(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}
QT_WARNING_POP
