/****************************************************************************
** Meta object code from reading C++ file 'PaneStack.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/view/PaneStack.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'PaneStack.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv9PaneStackE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::PaneStack::qt_create_metaobjectdata<qt_meta_tag_ZN2sv9PaneStackE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::PaneStack",
        "currentPaneChanged",
        "",
        "Pane*",
        "pane",
        "currentLayerChanged",
        "Layer*",
        "layer",
        "paneRightButtonMenuRequested",
        "QPoint",
        "position",
        "panePropertiesRightButtonMenuRequested",
        "layerPropertiesRightButtonMenuRequested",
        "propertyStacksResized",
        "width",
        "contextHelpChanged",
        "paneAdded",
        "paneHidden",
        "paneAboutToBeDeleted",
        "paneDeleted",
        "dropAccepted",
        "uriList",
        "text",
        "paneDeleteButtonClicked",
        "doubleClickSelectInvoked",
        "sv_frame_t",
        "frame",
        "propertyContainerAdded",
        "PropertyContainer*",
        "propertyContainerRemoved",
        "propertyContainerSelected",
        "View*",
        "client",
        "propertyContainerContextMenuRequested",
        "viewSelected",
        "v",
        "paneInteractedWith",
        "rightButtonMenuRequested",
        "paneDropAccepted",
        "indicatorClicked"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'currentPaneChanged'
        QtMocHelpers::SignalData<void(Pane *)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'currentLayerChanged'
        QtMocHelpers::SignalData<void(Pane *, Layer *)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 6, 7 },
        }}),
        // Signal 'paneRightButtonMenuRequested'
        QtMocHelpers::SignalData<void(Pane *, QPoint)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { 0x80000000 | 9, 10 },
        }}),
        // Signal 'panePropertiesRightButtonMenuRequested'
        QtMocHelpers::SignalData<void(Pane *, QPoint)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 }, { 0x80000000 | 9, 2 },
        }}),
        // Signal 'layerPropertiesRightButtonMenuRequested'
        QtMocHelpers::SignalData<void(Pane *, Layer *, QPoint)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 }, { 0x80000000 | 6, 2 }, { 0x80000000 | 9, 2 },
        }}),
        // Signal 'propertyStacksResized'
        QtMocHelpers::SignalData<void(int)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 14 },
        }}),
        // Signal 'propertyStacksResized'
        QtMocHelpers::SignalData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'contextHelpChanged'
        QtMocHelpers::SignalData<void(const QString &)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Signal 'paneAdded'
        QtMocHelpers::SignalData<void(Pane *)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'paneAdded'
        QtMocHelpers::SignalData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'paneHidden'
        QtMocHelpers::SignalData<void(Pane *)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'paneHidden'
        QtMocHelpers::SignalData<void()>(17, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'paneAboutToBeDeleted'
        QtMocHelpers::SignalData<void(Pane *)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'paneDeleted'
        QtMocHelpers::SignalData<void()>(19, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'dropAccepted'
        QtMocHelpers::SignalData<void(Pane *, QStringList)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { QMetaType::QStringList, 21 },
        }}),
        // Signal 'dropAccepted'
        QtMocHelpers::SignalData<void(Pane *, QString)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 }, { QMetaType::QString, 22 },
        }}),
        // Signal 'paneDeleteButtonClicked'
        QtMocHelpers::SignalData<void(Pane *)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 4 },
        }}),
        // Signal 'doubleClickSelectInvoked'
        QtMocHelpers::SignalData<void(sv_frame_t)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 25, 26 },
        }}),
        // Slot 'propertyContainerAdded'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(27, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 28, 2 },
        }}),
        // Slot 'propertyContainerRemoved'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(29, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 28, 2 },
        }}),
        // Slot 'propertyContainerSelected'
        QtMocHelpers::SlotData<void(View *, PropertyContainer *)>(30, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 32 }, { 0x80000000 | 28, 2 },
        }}),
        // Slot 'propertyContainerContextMenuRequested'
        QtMocHelpers::SlotData<void(View *, PropertyContainer *, QPoint)>(33, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 2 }, { 0x80000000 | 28, 2 }, { 0x80000000 | 9, 2 },
        }}),
        // Slot 'viewSelected'
        QtMocHelpers::SlotData<void(View *)>(34, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 35 },
        }}),
        // Slot 'paneInteractedWith'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'rightButtonMenuRequested'
        QtMocHelpers::SlotData<void(QPoint)>(37, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 2 },
        }}),
        // Slot 'paneDropAccepted'
        QtMocHelpers::SlotData<void(QStringList)>(38, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QStringList, 2 },
        }}),
        // Slot 'paneDropAccepted'
        QtMocHelpers::SlotData<void(QString)>(38, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'paneDeleteButtonClicked'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'indicatorClicked'
        QtMocHelpers::SlotData<void()>(39, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PaneStack, qt_meta_tag_ZN2sv9PaneStackE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::PaneStack::staticMetaObject = { {
    QMetaObject::SuperData::link<QFrame::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv9PaneStackE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv9PaneStackE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv9PaneStackE_t>.metaTypes,
    nullptr
} };

void sv::PaneStack::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PaneStack *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->currentPaneChanged((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 1: _t->currentLayerChanged((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[2]))); break;
        case 2: _t->paneRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[2]))); break;
        case 3: _t->panePropertiesRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[2]))); break;
        case 4: _t->layerPropertiesRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[3]))); break;
        case 5: _t->propertyStacksResized((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 6: _t->propertyStacksResized(); break;
        case 7: _t->contextHelpChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 8: _t->paneAdded((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 9: _t->paneAdded(); break;
        case 10: _t->paneHidden((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 11: _t->paneHidden(); break;
        case 12: _t->paneAboutToBeDeleted((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 13: _t->paneDeleted(); break;
        case 14: _t->dropAccepted((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[2]))); break;
        case 15: _t->dropAccepted((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 16: _t->paneDeleteButtonClicked((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 17: _t->doubleClickSelectInvoked((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 18: _t->propertyContainerAdded((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 19: _t->propertyContainerRemoved((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 20: _t->propertyContainerSelected((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[2]))); break;
        case 21: _t->propertyContainerContextMenuRequested((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[3]))); break;
        case 22: _t->viewSelected((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1]))); break;
        case 23: _t->paneInteractedWith(); break;
        case 24: _t->rightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 25: _t->paneDropAccepted((*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[1]))); break;
        case 26: _t->paneDropAccepted((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 27: _t->paneDeleteButtonClicked(); break;
        case 28: _t->indicatorClicked(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * )>(_a, &PaneStack::currentPaneChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * , Layer * )>(_a, &PaneStack::currentLayerChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * , QPoint )>(_a, &PaneStack::paneRightButtonMenuRequested, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * , QPoint )>(_a, &PaneStack::panePropertiesRightButtonMenuRequested, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * , Layer * , QPoint )>(_a, &PaneStack::layerPropertiesRightButtonMenuRequested, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(int )>(_a, &PaneStack::propertyStacksResized, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)()>(_a, &PaneStack::propertyStacksResized, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(const QString & )>(_a, &PaneStack::contextHelpChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * )>(_a, &PaneStack::paneAdded, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)()>(_a, &PaneStack::paneAdded, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * )>(_a, &PaneStack::paneHidden, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)()>(_a, &PaneStack::paneHidden, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * )>(_a, &PaneStack::paneAboutToBeDeleted, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)()>(_a, &PaneStack::paneDeleted, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * , QStringList )>(_a, &PaneStack::dropAccepted, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * , QString )>(_a, &PaneStack::dropAccepted, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(Pane * )>(_a, &PaneStack::paneDeleteButtonClicked, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (PaneStack::*)(sv_frame_t )>(_a, &PaneStack::doubleClickSelectInvoked, 17))
            return;
    }
}

const QMetaObject *sv::PaneStack::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::PaneStack::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv9PaneStackE_t>.strings))
        return static_cast<void*>(this);
    return QFrame::qt_metacast(_clname);
}

int sv::PaneStack::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QFrame::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 29)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 29;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 29)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 29;
    }
    return _id;
}

// SIGNAL 0
void sv::PaneStack::currentPaneChanged(Pane * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::PaneStack::currentLayerChanged(Pane * _t1, Layer * _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2);
}

// SIGNAL 2
void sv::PaneStack::paneRightButtonMenuRequested(Pane * _t1, QPoint _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1, _t2);
}

// SIGNAL 3
void sv::PaneStack::panePropertiesRightButtonMenuRequested(Pane * _t1, QPoint _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2);
}

// SIGNAL 4
void sv::PaneStack::layerPropertiesRightButtonMenuRequested(Pane * _t1, Layer * _t2, QPoint _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1, _t2, _t3);
}

// SIGNAL 5
void sv::PaneStack::propertyStacksResized(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void sv::PaneStack::propertyStacksResized()
{
    QMetaObject::activate(this, &staticMetaObject, 6, nullptr);
}

// SIGNAL 7
void sv::PaneStack::contextHelpChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void sv::PaneStack::paneAdded(Pane * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void sv::PaneStack::paneAdded()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void sv::PaneStack::paneHidden(Pane * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void sv::PaneStack::paneHidden()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}

// SIGNAL 12
void sv::PaneStack::paneAboutToBeDeleted(Pane * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1);
}

// SIGNAL 13
void sv::PaneStack::paneDeleted()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}

// SIGNAL 14
void sv::PaneStack::dropAccepted(Pane * _t1, QStringList _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1, _t2);
}

// SIGNAL 15
void sv::PaneStack::dropAccepted(Pane * _t1, QString _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1, _t2);
}

// SIGNAL 16
void sv::PaneStack::paneDeleteButtonClicked(Pane * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 16, nullptr, _t1);
}

// SIGNAL 17
void sv::PaneStack::doubleClickSelectInvoked(sv_frame_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 17, nullptr, _t1);
}
QT_WARNING_POP
