/****************************************************************************
** Meta object code from reading C++ file 'Document.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svapp/framework/Document.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'Document.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv8DocumentE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::Document::qt_create_metaobjectdata<qt_meta_tag_ZN2sv8DocumentE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::Document",
        "layerAdded",
        "",
        "Layer*",
        "layerRemoved",
        "layerAboutToBeDeleted",
        "layerInAView",
        "modelAdded",
        "ModelId",
        "mainModelChanged",
        "modelGenerationFailed",
        "transformName",
        "message",
        "modelGenerationWarning",
        "modelRegenerationFailed",
        "layerName",
        "modelRegenerationWarning",
        "alignmentComplete",
        "alignmentFailed",
        "activity",
        "performDeferredAlignment"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'layerAdded'
        QtMocHelpers::SignalData<void(Layer *)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'layerRemoved'
        QtMocHelpers::SignalData<void(Layer *)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'layerAboutToBeDeleted'
        QtMocHelpers::SignalData<void(Layer *)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'layerInAView'
        QtMocHelpers::SignalData<void(Layer *, bool)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Signal 'modelAdded'
        QtMocHelpers::SignalData<void(ModelId)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 2 },
        }}),
        // Signal 'mainModelChanged'
        QtMocHelpers::SignalData<void(ModelId)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 2 },
        }}),
        // Signal 'modelGenerationFailed'
        QtMocHelpers::SignalData<void(QString, QString)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 11 }, { QMetaType::QString, 12 },
        }}),
        // Signal 'modelGenerationWarning'
        QtMocHelpers::SignalData<void(QString, QString)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 11 }, { QMetaType::QString, 12 },
        }}),
        // Signal 'modelRegenerationFailed'
        QtMocHelpers::SignalData<void(QString, QString, QString)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 15 }, { QMetaType::QString, 11 }, { QMetaType::QString, 12 },
        }}),
        // Signal 'modelRegenerationWarning'
        QtMocHelpers::SignalData<void(QString, QString, QString)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 15 }, { QMetaType::QString, 11 }, { QMetaType::QString, 12 },
        }}),
        // Signal 'alignmentComplete'
        QtMocHelpers::SignalData<void(ModelId)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 2 },
        }}),
        // Signal 'alignmentFailed'
        QtMocHelpers::SignalData<void(ModelId, QString)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 8, 2 }, { QMetaType::QString, 12 },
        }}),
        // Signal 'activity'
        QtMocHelpers::SignalData<void(QString)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'performDeferredAlignment'
        QtMocHelpers::SlotData<void(ModelId)>(20, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 8, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<Document, qt_meta_tag_ZN2sv8DocumentE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::Document::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv8DocumentE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv8DocumentE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv8DocumentE_t>.metaTypes,
    nullptr
} };

void sv::Document::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<Document *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->layerAdded((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1]))); break;
        case 1: _t->layerRemoved((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1]))); break;
        case 2: _t->layerAboutToBeDeleted((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1]))); break;
        case 3: _t->layerInAView((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 4: _t->modelAdded((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 5: _t->mainModelChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 6: _t->modelGenerationFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 7: _t->modelGenerationWarning((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 8: _t->modelRegenerationFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 9: _t->modelRegenerationWarning((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 10: _t->alignmentComplete((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 11: _t->alignmentFailed((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 12: _t->activity((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 13: _t->performDeferredAlignment((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (Document::*)(Layer * )>(_a, &Document::layerAdded, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(Layer * )>(_a, &Document::layerRemoved, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(Layer * )>(_a, &Document::layerAboutToBeDeleted, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(Layer * , bool )>(_a, &Document::layerInAView, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(ModelId )>(_a, &Document::modelAdded, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(ModelId )>(_a, &Document::mainModelChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(QString , QString )>(_a, &Document::modelGenerationFailed, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(QString , QString )>(_a, &Document::modelGenerationWarning, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(QString , QString , QString )>(_a, &Document::modelRegenerationFailed, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(QString , QString , QString )>(_a, &Document::modelRegenerationWarning, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(ModelId )>(_a, &Document::alignmentComplete, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(ModelId , QString )>(_a, &Document::alignmentFailed, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (Document::*)(QString )>(_a, &Document::activity, 12))
            return;
    }
}

const QMetaObject *sv::Document::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::Document::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv8DocumentE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "XmlExportable"))
        return static_cast< XmlExportable*>(this);
    return QObject::qt_metacast(_clname);
}

int sv::Document::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 14)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 14;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 14)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 14;
    }
    return _id;
}

// SIGNAL 0
void sv::Document::layerAdded(Layer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::Document::layerRemoved(Layer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void sv::Document::layerAboutToBeDeleted(Layer * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::Document::layerInAView(Layer * _t1, bool _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1, _t2);
}

// SIGNAL 4
void sv::Document::modelAdded(ModelId _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void sv::Document::mainModelChanged(ModelId _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void sv::Document::modelGenerationFailed(QString _t1, QString _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1, _t2);
}

// SIGNAL 7
void sv::Document::modelGenerationWarning(QString _t1, QString _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1, _t2);
}

// SIGNAL 8
void sv::Document::modelRegenerationFailed(QString _t1, QString _t2, QString _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1, _t2, _t3);
}

// SIGNAL 9
void sv::Document::modelRegenerationWarning(QString _t1, QString _t2, QString _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1, _t2, _t3);
}

// SIGNAL 10
void sv::Document::alignmentComplete(ModelId _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void sv::Document::alignmentFailed(ModelId _t1, QString _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1, _t2);
}

// SIGNAL 12
void sv::Document::activity(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1);
}
QT_WARNING_POP
