/****************************************************************************
** Meta object code from reading C++ file 'Thumbwheel.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/widgets/Thumbwheel.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'Thumbwheel.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv10ThumbwheelE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::Thumbwheel::qt_create_metaobjectdata<qt_meta_tag_ZN2sv10ThumbwheelE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::Thumbwheel",
        "valueChanged",
        "",
        "mouseEntered",
        "mouseLeft",
        "setMinimumValue",
        "min",
        "setMaximumValue",
        "max",
        "setDefaultValue",
        "deft",
        "setSpeed",
        "speed",
        "setTracking",
        "tracking",
        "setShowScale",
        "show",
        "setValue",
        "value",
        "setMappedValue",
        "mappedValue",
        "scroll",
        "up",
        "resetToDefault",
        "edit",
        "updateMappedValue",
        "updateTitle",
        "contextMenuRequested",
        "QPoint"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'valueChanged'
        QtMocHelpers::SignalData<void(int)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Signal 'mouseEntered'
        QtMocHelpers::SignalData<void()>(3, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'mouseLeft'
        QtMocHelpers::SignalData<void()>(4, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'setMinimumValue'
        QtMocHelpers::SlotData<void(int)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 6 },
        }}),
        // Slot 'setMaximumValue'
        QtMocHelpers::SlotData<void(int)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 8 },
        }}),
        // Slot 'setDefaultValue'
        QtMocHelpers::SlotData<void(int)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Slot 'setSpeed'
        QtMocHelpers::SlotData<void(float)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 12 },
        }}),
        // Slot 'setTracking'
        QtMocHelpers::SlotData<void(bool)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 14 },
        }}),
        // Slot 'setShowScale'
        QtMocHelpers::SlotData<void(bool)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 16 },
        }}),
        // Slot 'setValue'
        QtMocHelpers::SlotData<void(int)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Slot 'setMappedValue'
        QtMocHelpers::SlotData<void(double)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Double, 20 },
        }}),
        // Slot 'scroll'
        QtMocHelpers::SlotData<void(bool)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 22 },
        }}),
        // Slot 'resetToDefault'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'edit'
        QtMocHelpers::SlotData<void()>(24, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'updateMappedValue'
        QtMocHelpers::SlotData<void(int)>(25, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 18 },
        }}),
        // Slot 'updateTitle'
        QtMocHelpers::SlotData<void()>(26, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'contextMenuRequested'
        QtMocHelpers::SlotData<void(const QPoint &)>(27, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 28, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<Thumbwheel, qt_meta_tag_ZN2sv10ThumbwheelE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::Thumbwheel::staticMetaObject = { {
    QMetaObject::SuperData::link<QWidget::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv10ThumbwheelE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv10ThumbwheelE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv10ThumbwheelE_t>.metaTypes,
    nullptr
} };

void sv::Thumbwheel::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<Thumbwheel *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->valueChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 1: _t->mouseEntered(); break;
        case 2: _t->mouseLeft(); break;
        case 3: _t->setMinimumValue((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 4: _t->setMaximumValue((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 5: _t->setDefaultValue((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 6: _t->setSpeed((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 7: _t->setTracking((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 8: _t->setShowScale((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 9: _t->setValue((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 10: _t->setMappedValue((*reinterpret_cast<std::add_pointer_t<double>>(_a[1]))); break;
        case 11: _t->scroll((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 12: _t->resetToDefault(); break;
        case 13: _t->edit(); break;
        case 14: _t->updateMappedValue((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 15: _t->updateTitle(); break;
        case 16: _t->contextMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (Thumbwheel::*)(int )>(_a, &Thumbwheel::valueChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (Thumbwheel::*)()>(_a, &Thumbwheel::mouseEntered, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (Thumbwheel::*)()>(_a, &Thumbwheel::mouseLeft, 2))
            return;
    }
}

const QMetaObject *sv::Thumbwheel::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::Thumbwheel::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv10ThumbwheelE_t>.strings))
        return static_cast<void*>(this);
    return QWidget::qt_metacast(_clname);
}

int sv::Thumbwheel::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QWidget::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 17)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 17;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 17)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 17;
    }
    return _id;
}

// SIGNAL 0
void sv::Thumbwheel::valueChanged(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::Thumbwheel::mouseEntered()
{
    QMetaObject::activate(this, &staticMetaObject, 1, nullptr);
}

// SIGNAL 2
void sv::Thumbwheel::mouseLeft()
{
    QMetaObject::activate(this, &staticMetaObject, 2, nullptr);
}
QT_WARNING_POP
