/****************************************************************************
** Meta object code from reading C++ file 'MainWindow.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/main/MainWindow.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'MainWindow.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN10MainWindowE_t {};
} // unnamed namespace

template <> constexpr inline auto MainWindow::qt_create_metaobjectdata<qt_meta_tag_ZN10MainWindowE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "MainWindow",
        "canChangeSolo",
        "",
        "canAlign",
        "preferenceChanged",
        "sv::PropertyContainer::PropertyName",
        "coloursChanged",
        "commitData",
        "mayAskUser",
        "goFullScreen",
        "endFullScreen",
        "importAudio",
        "importMoreAudio",
        "replaceMainAudio",
        "openSomething",
        "openLocation",
        "openRecentFile",
        "applyTemplate",
        "exportAudio",
        "exportAudioData",
        "convertAudio",
        "importLayer",
        "exportLayer",
        "exportImage",
        "exportSVG",
        "browseRecordedAudio",
        "saveSession",
        "saveSessionAs",
        "newSession",
        "closeSession",
        "preferences",
        "sampleRateMismatch",
        "sv::sv_samplerate_t",
        "audioOverloadPluginDisabled",
        "toolNavigateSelected",
        "toolSelectSelected",
        "toolEditSelected",
        "toolDrawSelected",
        "toolEraseSelected",
        "toolMeasureSelected",
        "documentModified",
        "documentRestored",
        "documentReplaced",
        "updateMenuStates",
        "updateDescriptionLabel",
        "setInstantsNumbering",
        "setInstantsCounterCycle",
        "setInstantsCounters",
        "resetInstantsCounters",
        "subdivideInstants",
        "winnowInstants",
        "modelGenerationFailed",
        "modelGenerationWarning",
        "modelRegenerationFailed",
        "modelRegenerationWarning",
        "alignmentFailed",
        "sv::ModelId",
        "paneRightButtonMenuRequested",
        "sv::Pane*",
        "QPoint",
        "point",
        "panePropertiesRightButtonMenuRequested",
        "layerPropertiesRightButtonMenuRequested",
        "sv::Layer*",
        "propertyStacksResized",
        "addPane",
        "addLayer",
        "transformId",
        "renameCurrentLayer",
        "findTransform",
        "paneAdded",
        "paneHidden",
        "paneAboutToBeDeleted",
        "paneDropAccepted",
        "paneCancelButtonPressed",
        "setupRecentFilesMenu",
        "setupRecentTransformsMenu",
        "setupTemplatesMenu",
        "playSpeedChanged",
        "playSoloToggled",
        "alignToggled",
        "currentPaneChanged",
        "speedUpPlayback",
        "slowDownPlayback",
        "restoreNormalPlayback",
        "monitoringLevelsChanged",
        "layerRemoved",
        "layerInAView",
        "mainModelChanged",
        "mainModelGainChanged",
        "mainModelPanChanged",
        "modelAdded",
        "showLayerTree",
        "showActivityLog",
        "showUnitConverter",
        "mouseEnteredWidget",
        "mouseLeftWidget",
        "handleOSCMessage",
        "sv::OSCMessage",
        "midiEventsAvailable",
        "playStatusChanged",
        "installedTransformsPopulated",
        "populateTransformsMenu",
        "betaReleaseWarning",
        "pluginPopulationWarning",
        "text",
        "saveSessionAsTemplate",
        "manageSavedTemplates",
        "website",
        "help",
        "about",
        "whatsNew",
        "keyReference",
        "newerVersionAvailable"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'canChangeSolo'
        QtMocHelpers::SignalData<void(bool)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canAlign'
        QtMocHelpers::SignalData<void(bool)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'preferenceChanged'
        QtMocHelpers::SlotData<void(sv::PropertyContainer::PropertyName)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 2 },
        }}),
        // Slot 'coloursChanged'
        QtMocHelpers::SlotData<void()>(6, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'commitData'
        QtMocHelpers::SlotData<bool(bool)>(7, 2, QMC::AccessPublic, QMetaType::Bool, {{
            { QMetaType::Bool, 8 },
        }}),
        // Slot 'goFullScreen'
        QtMocHelpers::SlotData<void()>(9, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'endFullScreen'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'importAudio'
        QtMocHelpers::SlotData<void()>(11, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'importMoreAudio'
        QtMocHelpers::SlotData<void()>(12, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'replaceMainAudio'
        QtMocHelpers::SlotData<void()>(13, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'openSomething'
        QtMocHelpers::SlotData<void()>(14, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'openLocation'
        QtMocHelpers::SlotData<void()>(15, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'openRecentFile'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'applyTemplate'
        QtMocHelpers::SlotData<void()>(17, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'exportAudio'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'exportAudioData'
        QtMocHelpers::SlotData<void()>(19, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'convertAudio'
        QtMocHelpers::SlotData<void()>(20, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'importLayer'
        QtMocHelpers::SlotData<void()>(21, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'exportLayer'
        QtMocHelpers::SlotData<void()>(22, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'exportImage'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'exportSVG'
        QtMocHelpers::SlotData<void()>(24, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'browseRecordedAudio'
        QtMocHelpers::SlotData<void()>(25, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'saveSession'
        QtMocHelpers::SlotData<void()>(26, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'saveSessionAs'
        QtMocHelpers::SlotData<void()>(27, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'newSession'
        QtMocHelpers::SlotData<void()>(28, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'closeSession'
        QtMocHelpers::SlotData<void()>(29, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'preferences'
        QtMocHelpers::SlotData<void()>(30, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'sampleRateMismatch'
        QtMocHelpers::SlotData<void(sv::sv_samplerate_t, sv::sv_samplerate_t, bool)>(31, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 32, 2 }, { 0x80000000 | 32, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Slot 'audioOverloadPluginDisabled'
        QtMocHelpers::SlotData<void()>(33, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toolNavigateSelected'
        QtMocHelpers::SlotData<void()>(34, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toolSelectSelected'
        QtMocHelpers::SlotData<void()>(35, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toolEditSelected'
        QtMocHelpers::SlotData<void()>(36, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toolDrawSelected'
        QtMocHelpers::SlotData<void()>(37, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toolEraseSelected'
        QtMocHelpers::SlotData<void()>(38, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toolMeasureSelected'
        QtMocHelpers::SlotData<void()>(39, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'documentModified'
        QtMocHelpers::SlotData<void()>(40, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'documentRestored'
        QtMocHelpers::SlotData<void()>(41, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'documentReplaced'
        QtMocHelpers::SlotData<void()>(42, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'updateMenuStates'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'updateDescriptionLabel'
        QtMocHelpers::SlotData<void()>(44, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'setInstantsNumbering'
        QtMocHelpers::SlotData<void()>(45, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'setInstantsCounterCycle'
        QtMocHelpers::SlotData<void()>(46, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'setInstantsCounters'
        QtMocHelpers::SlotData<void()>(47, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'resetInstantsCounters'
        QtMocHelpers::SlotData<void()>(48, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'subdivideInstants'
        QtMocHelpers::SlotData<void()>(49, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'winnowInstants'
        QtMocHelpers::SlotData<void()>(50, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'modelGenerationFailed'
        QtMocHelpers::SlotData<void(QString, QString)>(51, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'modelGenerationWarning'
        QtMocHelpers::SlotData<void(QString, QString)>(52, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'modelRegenerationFailed'
        QtMocHelpers::SlotData<void(QString, QString, QString)>(53, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'modelRegenerationWarning'
        QtMocHelpers::SlotData<void(QString, QString, QString)>(54, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'alignmentFailed'
        QtMocHelpers::SlotData<void(sv::ModelId, QString)>(55, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 56, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'paneRightButtonMenuRequested'
        QtMocHelpers::SlotData<void(sv::Pane *, QPoint)>(57, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 }, { 0x80000000 | 59, 60 },
        }}),
        // Slot 'panePropertiesRightButtonMenuRequested'
        QtMocHelpers::SlotData<void(sv::Pane *, QPoint)>(61, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 }, { 0x80000000 | 59, 60 },
        }}),
        // Slot 'layerPropertiesRightButtonMenuRequested'
        QtMocHelpers::SlotData<void(sv::Pane *, sv::Layer *, QPoint)>(62, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 }, { 0x80000000 | 63, 2 }, { 0x80000000 | 59, 60 },
        }}),
        // Slot 'propertyStacksResized'
        QtMocHelpers::SlotData<void(int)>(64, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'addPane'
        QtMocHelpers::SlotData<void()>(65, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'addLayer'
        QtMocHelpers::SlotData<void()>(66, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'addLayer'
        QtMocHelpers::SlotData<void(QString)>(66, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 67 },
        }}),
        // Slot 'renameCurrentLayer'
        QtMocHelpers::SlotData<void()>(68, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'findTransform'
        QtMocHelpers::SlotData<void()>(69, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'paneAdded'
        QtMocHelpers::SlotData<void(sv::Pane *)>(70, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 },
        }}),
        // Slot 'paneHidden'
        QtMocHelpers::SlotData<void(sv::Pane *)>(71, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 },
        }}),
        // Slot 'paneAboutToBeDeleted'
        QtMocHelpers::SlotData<void(sv::Pane *)>(72, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 },
        }}),
        // Slot 'paneDropAccepted'
        QtMocHelpers::SlotData<void(sv::Pane *, QStringList)>(73, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 }, { QMetaType::QStringList, 2 },
        }}),
        // Slot 'paneDropAccepted'
        QtMocHelpers::SlotData<void(sv::Pane *, QString)>(73, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'paneCancelButtonPressed'
        QtMocHelpers::SlotData<void(sv::Layer *)>(74, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 63, 2 },
        }}),
        // Slot 'setupRecentFilesMenu'
        QtMocHelpers::SlotData<void()>(75, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'setupRecentTransformsMenu'
        QtMocHelpers::SlotData<void()>(76, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'setupTemplatesMenu'
        QtMocHelpers::SlotData<void()>(77, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playSpeedChanged'
        QtMocHelpers::SlotData<void(int)>(78, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'playSoloToggled'
        QtMocHelpers::SlotData<void()>(79, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'alignToggled'
        QtMocHelpers::SlotData<void()>(80, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'currentPaneChanged'
        QtMocHelpers::SlotData<void(sv::Pane *)>(81, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 58, 2 },
        }}),
        // Slot 'speedUpPlayback'
        QtMocHelpers::SlotData<void()>(82, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'slowDownPlayback'
        QtMocHelpers::SlotData<void()>(83, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'restoreNormalPlayback'
        QtMocHelpers::SlotData<void()>(84, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'monitoringLevelsChanged'
        QtMocHelpers::SlotData<void(float, float)>(85, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Float, 2 }, { QMetaType::Float, 2 },
        }}),
        // Slot 'layerRemoved'
        QtMocHelpers::SlotData<void(sv::Layer *)>(86, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 63, 2 },
        }}),
        // Slot 'layerInAView'
        QtMocHelpers::SlotData<void(sv::Layer *, bool)>(87, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 63, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Slot 'mainModelChanged'
        QtMocHelpers::SlotData<void(sv::ModelId)>(88, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 56, 2 },
        }}),
        // Slot 'mainModelGainChanged'
        QtMocHelpers::SlotData<void(float)>(89, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Float, 2 },
        }}),
        // Slot 'mainModelPanChanged'
        QtMocHelpers::SlotData<void(float)>(90, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Float, 2 },
        }}),
        // Slot 'modelAdded'
        QtMocHelpers::SlotData<void(sv::ModelId)>(91, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 56, 2 },
        }}),
        // Slot 'showLayerTree'
        QtMocHelpers::SlotData<void()>(92, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'showActivityLog'
        QtMocHelpers::SlotData<void()>(93, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'showUnitConverter'
        QtMocHelpers::SlotData<void()>(94, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'mouseEnteredWidget'
        QtMocHelpers::SlotData<void()>(95, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'mouseLeftWidget'
        QtMocHelpers::SlotData<void()>(96, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'handleOSCMessage'
        QtMocHelpers::SlotData<void(const sv::OSCMessage &)>(97, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 98, 2 },
        }}),
        // Slot 'midiEventsAvailable'
        QtMocHelpers::SlotData<void()>(99, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playStatusChanged'
        QtMocHelpers::SlotData<void(bool)>(100, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'installedTransformsPopulated'
        QtMocHelpers::SlotData<void()>(101, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'populateTransformsMenu'
        QtMocHelpers::SlotData<void()>(102, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'betaReleaseWarning'
        QtMocHelpers::SlotData<void()>(103, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'pluginPopulationWarning'
        QtMocHelpers::SlotData<void(QString)>(104, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 105 },
        }}),
        // Slot 'saveSessionAsTemplate'
        QtMocHelpers::SlotData<void()>(106, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'manageSavedTemplates'
        QtMocHelpers::SlotData<void()>(107, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'website'
        QtMocHelpers::SlotData<void()>(108, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'help'
        QtMocHelpers::SlotData<void()>(109, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'about'
        QtMocHelpers::SlotData<void()>(110, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'whatsNew'
        QtMocHelpers::SlotData<void()>(111, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'keyReference'
        QtMocHelpers::SlotData<void()>(112, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'newerVersionAvailable'
        QtMocHelpers::SlotData<void(QString)>(113, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<MainWindow, qt_meta_tag_ZN10MainWindowE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject MainWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<sv::MainWindowBase::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN10MainWindowE_t>.metaTypes,
    nullptr
} };

void MainWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<MainWindow *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->canChangeSolo((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 1: _t->canAlign((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 2: _t->preferenceChanged((*reinterpret_cast<std::add_pointer_t<sv::PropertyContainer::PropertyName>>(_a[1]))); break;
        case 3: _t->coloursChanged(); break;
        case 4: { bool _r = _t->commitData((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1])));
            if (_a[0]) *reinterpret_cast<bool*>(_a[0]) = std::move(_r); }  break;
        case 5: _t->goFullScreen(); break;
        case 6: _t->endFullScreen(); break;
        case 7: _t->importAudio(); break;
        case 8: _t->importMoreAudio(); break;
        case 9: _t->replaceMainAudio(); break;
        case 10: _t->openSomething(); break;
        case 11: _t->openLocation(); break;
        case 12: _t->openRecentFile(); break;
        case 13: _t->applyTemplate(); break;
        case 14: _t->exportAudio(); break;
        case 15: _t->exportAudioData(); break;
        case 16: _t->convertAudio(); break;
        case 17: _t->importLayer(); break;
        case 18: _t->exportLayer(); break;
        case 19: _t->exportImage(); break;
        case 20: _t->exportSVG(); break;
        case 21: _t->browseRecordedAudio(); break;
        case 22: _t->saveSession(); break;
        case 23: _t->saveSessionAs(); break;
        case 24: _t->newSession(); break;
        case 25: _t->closeSession(); break;
        case 26: _t->preferences(); break;
        case 27: _t->sampleRateMismatch((*reinterpret_cast<std::add_pointer_t<sv::sv_samplerate_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv::sv_samplerate_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 28: _t->audioOverloadPluginDisabled(); break;
        case 29: _t->toolNavigateSelected(); break;
        case 30: _t->toolSelectSelected(); break;
        case 31: _t->toolEditSelected(); break;
        case 32: _t->toolDrawSelected(); break;
        case 33: _t->toolEraseSelected(); break;
        case 34: _t->toolMeasureSelected(); break;
        case 35: _t->documentModified(); break;
        case 36: _t->documentRestored(); break;
        case 37: _t->documentReplaced(); break;
        case 38: _t->updateMenuStates(); break;
        case 39: _t->updateDescriptionLabel(); break;
        case 40: _t->setInstantsNumbering(); break;
        case 41: _t->setInstantsCounterCycle(); break;
        case 42: _t->setInstantsCounters(); break;
        case 43: _t->resetInstantsCounters(); break;
        case 44: _t->subdivideInstants(); break;
        case 45: _t->winnowInstants(); break;
        case 46: _t->modelGenerationFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 47: _t->modelGenerationWarning((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 48: _t->modelRegenerationFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 49: _t->modelRegenerationWarning((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 50: _t->alignmentFailed((*reinterpret_cast<std::add_pointer_t<sv::ModelId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 51: _t->paneRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[2]))); break;
        case 52: _t->panePropertiesRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[2]))); break;
        case 53: _t->layerPropertiesRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv::Layer*>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[3]))); break;
        case 54: _t->propertyStacksResized((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 55: _t->addPane(); break;
        case 56: _t->addLayer(); break;
        case 57: _t->addLayer((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 58: _t->renameCurrentLayer(); break;
        case 59: _t->findTransform(); break;
        case 60: _t->paneAdded((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1]))); break;
        case 61: _t->paneHidden((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1]))); break;
        case 62: _t->paneAboutToBeDeleted((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1]))); break;
        case 63: _t->paneDropAccepted((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[2]))); break;
        case 64: _t->paneDropAccepted((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 65: _t->paneCancelButtonPressed((*reinterpret_cast<std::add_pointer_t<sv::Layer*>>(_a[1]))); break;
        case 66: _t->setupRecentFilesMenu(); break;
        case 67: _t->setupRecentTransformsMenu(); break;
        case 68: _t->setupTemplatesMenu(); break;
        case 69: _t->playSpeedChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 70: _t->playSoloToggled(); break;
        case 71: _t->alignToggled(); break;
        case 72: _t->currentPaneChanged((*reinterpret_cast<std::add_pointer_t<sv::Pane*>>(_a[1]))); break;
        case 73: _t->speedUpPlayback(); break;
        case 74: _t->slowDownPlayback(); break;
        case 75: _t->restoreNormalPlayback(); break;
        case 76: _t->monitoringLevelsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[2]))); break;
        case 77: _t->layerRemoved((*reinterpret_cast<std::add_pointer_t<sv::Layer*>>(_a[1]))); break;
        case 78: _t->layerInAView((*reinterpret_cast<std::add_pointer_t<sv::Layer*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 79: _t->mainModelChanged((*reinterpret_cast<std::add_pointer_t<sv::ModelId>>(_a[1]))); break;
        case 80: _t->mainModelGainChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 81: _t->mainModelPanChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 82: _t->modelAdded((*reinterpret_cast<std::add_pointer_t<sv::ModelId>>(_a[1]))); break;
        case 83: _t->showLayerTree(); break;
        case 84: _t->showActivityLog(); break;
        case 85: _t->showUnitConverter(); break;
        case 86: _t->mouseEnteredWidget(); break;
        case 87: _t->mouseLeftWidget(); break;
        case 88: _t->handleOSCMessage((*reinterpret_cast<std::add_pointer_t<sv::OSCMessage>>(_a[1]))); break;
        case 89: _t->midiEventsAvailable(); break;
        case 90: _t->playStatusChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 91: _t->installedTransformsPopulated(); break;
        case 92: _t->populateTransformsMenu(); break;
        case 93: _t->betaReleaseWarning(); break;
        case 94: _t->pluginPopulationWarning((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 95: _t->saveSessionAsTemplate(); break;
        case 96: _t->manageSavedTemplates(); break;
        case 97: _t->website(); break;
        case 98: _t->help(); break;
        case 99: _t->about(); break;
        case 100: _t->whatsNew(); break;
        case 101: _t->keyReference(); break;
        case 102: _t->newerVersionAvailable((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(bool )>(_a, &MainWindow::canChangeSolo, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindow::*)(bool )>(_a, &MainWindow::canAlign, 1))
            return;
    }
}

const QMetaObject *MainWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *MainWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN10MainWindowE_t>.strings))
        return static_cast<void*>(this);
    return sv::MainWindowBase::qt_metacast(_clname);
}

int MainWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = sv::MainWindowBase::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 103)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 103;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 103)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 103;
    }
    return _id;
}

// SIGNAL 0
void MainWindow::canChangeSolo(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void MainWindow::canAlign(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}
QT_WARNING_POP
