/****************************************************************************
** Meta object code from reading C++ file 'AudioCallbackPlaySource.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svapp/audio/AudioCallbackPlaySource.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'AudioCallbackPlaySource.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv23AudioCallbackPlaySourceE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::AudioCallbackPlaySource::qt_create_metaobjectdata<qt_meta_tag_ZN2sv23AudioCallbackPlaySourceE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::AudioCallbackPlaySource",
        "playStatusChanged",
        "",
        "isPlaying",
        "sampleRateMismatch",
        "sv_samplerate_t",
        "requested",
        "available",
        "willResample",
        "channelCountIncreased",
        "count",
        "audioOverloadPluginDisabled",
        "activity",
        "audioProcessingOverload",
        "selectionChanged",
        "playLoopModeChanged",
        "playSelectionModeChanged",
        "playParametersChanged",
        "preferenceChanged",
        "PropertyContainer::PropertyName",
        "modelChangedWithin",
        "ModelId",
        "sv_frame_t",
        "startFrame",
        "endFrame"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'playStatusChanged'
        QtMocHelpers::SignalData<void(bool)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 3 },
        }}),
        // Signal 'sampleRateMismatch'
        QtMocHelpers::SignalData<void(sv_samplerate_t, sv_samplerate_t, bool)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 6 }, { 0x80000000 | 5, 7 }, { QMetaType::Bool, 8 },
        }}),
        // Signal 'channelCountIncreased'
        QtMocHelpers::SignalData<void(int)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 10 },
        }}),
        // Signal 'audioOverloadPluginDisabled'
        QtMocHelpers::SignalData<void()>(11, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'activity'
        QtMocHelpers::SignalData<void(QString)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'audioProcessingOverload'
        QtMocHelpers::SlotData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'selectionChanged'
        QtMocHelpers::SlotData<void()>(14, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playLoopModeChanged'
        QtMocHelpers::SlotData<void()>(15, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playSelectionModeChanged'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playParametersChanged'
        QtMocHelpers::SlotData<void(int)>(17, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'preferenceChanged'
        QtMocHelpers::SlotData<void(PropertyContainer::PropertyName)>(18, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 19, 2 },
        }}),
        // Slot 'modelChangedWithin'
        QtMocHelpers::SlotData<void(ModelId, sv_frame_t, sv_frame_t)>(20, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 21, 2 }, { 0x80000000 | 22, 23 }, { 0x80000000 | 22, 24 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<AudioCallbackPlaySource, qt_meta_tag_ZN2sv23AudioCallbackPlaySourceE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::AudioCallbackPlaySource::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv23AudioCallbackPlaySourceE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv23AudioCallbackPlaySourceE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv23AudioCallbackPlaySourceE_t>.metaTypes,
    nullptr
} };

void sv::AudioCallbackPlaySource::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<AudioCallbackPlaySource *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->playStatusChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 1: _t->sampleRateMismatch((*reinterpret_cast<std::add_pointer_t<sv_samplerate_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_samplerate_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 2: _t->channelCountIncreased((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 3: _t->audioOverloadPluginDisabled(); break;
        case 4: _t->activity((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 5: _t->audioProcessingOverload(); break;
        case 6: _t->selectionChanged(); break;
        case 7: _t->playLoopModeChanged(); break;
        case 8: _t->playSelectionModeChanged(); break;
        case 9: _t->playParametersChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 10: _t->preferenceChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer::PropertyName>>(_a[1]))); break;
        case 11: _t->modelChangedWithin((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[3]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (AudioCallbackPlaySource::*)(bool )>(_a, &AudioCallbackPlaySource::playStatusChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioCallbackPlaySource::*)(sv_samplerate_t , sv_samplerate_t , bool )>(_a, &AudioCallbackPlaySource::sampleRateMismatch, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioCallbackPlaySource::*)(int )>(_a, &AudioCallbackPlaySource::channelCountIncreased, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioCallbackPlaySource::*)()>(_a, &AudioCallbackPlaySource::audioOverloadPluginDisabled, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (AudioCallbackPlaySource::*)(QString )>(_a, &AudioCallbackPlaySource::activity, 4))
            return;
    }
}

const QMetaObject *sv::AudioCallbackPlaySource::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::AudioCallbackPlaySource::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv23AudioCallbackPlaySourceE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "AudioPlaySource"))
        return static_cast< AudioPlaySource*>(this);
    if (!strcmp(_clname, "breakfastquay::ApplicationPlaybackSource"))
        return static_cast< breakfastquay::ApplicationPlaybackSource*>(this);
    return QObject::qt_metacast(_clname);
}

int sv::AudioCallbackPlaySource::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 12)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 12;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 12)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 12;
    }
    return _id;
}

// SIGNAL 0
void sv::AudioCallbackPlaySource::playStatusChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::AudioCallbackPlaySource::sampleRateMismatch(sv_samplerate_t _t1, sv_samplerate_t _t2, bool _t3)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1, _t2, _t3);
}

// SIGNAL 2
void sv::AudioCallbackPlaySource::channelCountIncreased(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::AudioCallbackPlaySource::audioOverloadPluginDisabled()
{
    QMetaObject::activate(this, &staticMetaObject, 3, nullptr);
}

// SIGNAL 4
void sv::AudioCallbackPlaySource::activity(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}
QT_WARNING_POP
