/****************************************************************************
** Meta object code from reading C++ file 'PropertyBox.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/widgets/PropertyBox.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'PropertyBox.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv11PropertyBoxE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::PropertyBox::qt_create_metaobjectdata<qt_meta_tag_ZN2sv11PropertyBoxE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::PropertyBox",
        "showLayer",
        "",
        "contextHelpChanged",
        "propertyContainerPropertyChanged",
        "PropertyContainer*",
        "propertyContainerPropertyRangeChanged",
        "playClipChanged",
        "layerVisibilityChanged",
        "propertyControllerChanged",
        "propertyControllerResetRequested",
        "playAudibleChanged",
        "playAudibleButtonChanged",
        "playGainControlChanged",
        "playPanControlChanged",
        "populateViewPlayFrame",
        "unitDatabaseChanged",
        "editPlayParameters",
        "mouseEnteredWidget",
        "mouseLeftWidget",
        "contextMenuRequested",
        "QPoint"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'showLayer'
        QtMocHelpers::SignalData<void(bool)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'contextHelpChanged'
        QtMocHelpers::SignalData<void(const QString &)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'propertyContainerPropertyChanged'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 2 },
        }}),
        // Slot 'propertyContainerPropertyRangeChanged'
        QtMocHelpers::SlotData<void(PropertyContainer *)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 5, 2 },
        }}),
        // Slot 'playClipChanged'
        QtMocHelpers::SlotData<void(QString)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'layerVisibilityChanged'
        QtMocHelpers::SlotData<void(bool)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'propertyControllerChanged'
        QtMocHelpers::SlotData<void(int)>(9, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'propertyControllerChanged'
        QtMocHelpers::SlotData<void(bool)>(9, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'propertyControllerResetRequested'
        QtMocHelpers::SlotData<void()>(10, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playAudibleChanged'
        QtMocHelpers::SlotData<void(bool)>(11, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'playAudibleButtonChanged'
        QtMocHelpers::SlotData<void(bool)>(12, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'playGainControlChanged'
        QtMocHelpers::SlotData<void(float)>(13, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Float, 2 },
        }}),
        // Slot 'playPanControlChanged'
        QtMocHelpers::SlotData<void(float)>(14, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Float, 2 },
        }}),
        // Slot 'populateViewPlayFrame'
        QtMocHelpers::SlotData<void()>(15, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'unitDatabaseChanged'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'editPlayParameters'
        QtMocHelpers::SlotData<void()>(17, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'mouseEnteredWidget'
        QtMocHelpers::SlotData<void()>(18, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'mouseLeftWidget'
        QtMocHelpers::SlotData<void()>(19, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'contextMenuRequested'
        QtMocHelpers::SlotData<void(const QPoint &)>(20, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 21, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<PropertyBox, qt_meta_tag_ZN2sv11PropertyBoxE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::PropertyBox::staticMetaObject = { {
    QMetaObject::SuperData::link<QFrame::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11PropertyBoxE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11PropertyBoxE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv11PropertyBoxE_t>.metaTypes,
    nullptr
} };

void sv::PropertyBox::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<PropertyBox *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->showLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 1: _t->contextHelpChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 2: _t->propertyContainerPropertyChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 3: _t->propertyContainerPropertyRangeChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[1]))); break;
        case 4: _t->playClipChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 5: _t->layerVisibilityChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 6: _t->propertyControllerChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 7: _t->propertyControllerChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 8: _t->propertyControllerResetRequested(); break;
        case 9: _t->playAudibleChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 10: _t->playAudibleButtonChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 11: _t->playGainControlChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 12: _t->playPanControlChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 13: _t->populateViewPlayFrame(); break;
        case 14: _t->unitDatabaseChanged(); break;
        case 15: _t->editPlayParameters(); break;
        case 16: _t->mouseEnteredWidget(); break;
        case 17: _t->mouseLeftWidget(); break;
        case 18: _t->contextMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (PropertyBox::*)(bool )>(_a, &PropertyBox::showLayer, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (PropertyBox::*)(const QString & )>(_a, &PropertyBox::contextHelpChanged, 1))
            return;
    }
}

const QMetaObject *sv::PropertyBox::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::PropertyBox::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv11PropertyBoxE_t>.strings))
        return static_cast<void*>(this);
    return QFrame::qt_metacast(_clname);
}

int sv::PropertyBox::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QFrame::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 19)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 19;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 19)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 19;
    }
    return _id;
}

// SIGNAL 0
void sv::PropertyBox::showLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::PropertyBox::contextHelpChanged(const QString & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}
QT_WARNING_POP
