/****************************************************************************
** Meta object code from reading C++ file 'MainWindowBase.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svapp/framework/MainWindowBase.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'MainWindowBase.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv14MainWindowBaseE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::MainWindowBase::qt_create_metaobjectdata<qt_meta_tag_ZN2sv14MainWindowBaseE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::MainWindowBase",
        "canAddPane",
        "",
        "canDeleteCurrentPane",
        "canAddLayer",
        "canImportMoreAudio",
        "canReplaceMainAudio",
        "canImportLayer",
        "canChangeSessionTemplate",
        "canExportAudio",
        "canExportLayer",
        "canExportImage",
        "canRenameLayer",
        "canEditLayer",
        "canEditLayerTabular",
        "canMeasureLayer",
        "canSelect",
        "canClearSelection",
        "canEditSelection",
        "canDeleteSelection",
        "canPaste",
        "canInsertInstant",
        "canInsertInstantsAtBoundaries",
        "canInsertItemAtSelection",
        "canRenumberInstants",
        "canSubdivideInstants",
        "canWinnowInstants",
        "canDeleteCurrentLayer",
        "canZoom",
        "canScroll",
        "canPlay",
        "canRecord",
        "canFfwd",
        "canRewind",
        "canPlaySelection",
        "canSpeedUpPlayback",
        "canSlowDownPlayback",
        "canChangePlaybackSpeed",
        "canSelectPreviousPane",
        "canSelectNextPane",
        "canSelectPreviousLayer",
        "canSelectNextLayer",
        "canSave",
        "canSaveAs",
        "hideSplash",
        "QWidget*",
        "sessionLoaded",
        "audioFileLoaded",
        "replacedDocument",
        "activity",
        "preferenceChanged",
        "PropertyContainer::PropertyName",
        "resizeConstrained",
        "QSize",
        "recreateAudioIO",
        "zoomIn",
        "zoomOut",
        "zoomToFit",
        "zoomDefault",
        "scrollLeft",
        "scrollRight",
        "jumpLeft",
        "jumpRight",
        "peekLeft",
        "peekRight",
        "showNoOverlays",
        "showMinimalOverlays",
        "showAllOverlays",
        "toggleTimeRulers",
        "toggleZoomWheels",
        "togglePropertyBoxes",
        "toggleStatusBar",
        "toggleCentreLine",
        "play",
        "ffwd",
        "ffwdEnd",
        "rewind",
        "rewindStart",
        "record",
        "stop",
        "ffwdSimilar",
        "rewindSimilar",
        "deleteCurrentPane",
        "deleteCurrentLayer",
        "editCurrentLayer",
        "previousPane",
        "nextPane",
        "previousLayer",
        "nextLayer",
        "playLoopToggled",
        "playSelectionToggled",
        "playSoloToggled",
        "audioChannelCountIncreased",
        "count",
        "sampleRateMismatch",
        "sv_samplerate_t",
        "audioOverloadPluginDisabled",
        "playbackFrameChanged",
        "sv_frame_t",
        "globalCentreFrameChanged",
        "viewCentreFrameChanged",
        "View*",
        "viewZoomLevelChanged",
        "ZoomLevel",
        "monitoringLevelsChanged",
        "recordDurationChanged",
        "currentPaneChanged",
        "Pane*",
        "currentLayerChanged",
        "Layer*",
        "selectAll",
        "selectToStart",
        "selectToEnd",
        "selectVisible",
        "clearSelection",
        "cut",
        "copy",
        "paste",
        "pasteAtPlaybackPosition",
        "pasteRelative",
        "offset",
        "deleteSelected",
        "insertInstant",
        "insertInstantAt",
        "insertInstantsAtBoundaries",
        "insertItemAtSelection",
        "insertItemAt",
        "renumberInstants",
        "subdivideInstantsBy",
        "winnowInstantsBy",
        "documentModified",
        "documentRestored",
        "layerAdded",
        "layerRemoved",
        "layerAboutToBeDeleted",
        "layerInAView",
        "mainModelChanged",
        "ModelId",
        "modelAdded",
        "updateMenuStates",
        "updateDescriptionLabel",
        "updateWindowTitle",
        "modelGenerationFailed",
        "modelGenerationWarning",
        "modelRegenerationFailed",
        "modelRegenerationWarning",
        "alignmentComplete",
        "alignmentFailed",
        "paneRightButtonMenuRequested",
        "QPoint",
        "point",
        "panePropertiesRightButtonMenuRequested",
        "layerPropertiesRightButtonMenuRequested",
        "paneAdded",
        "paneHidden",
        "paneAboutToBeDeleted",
        "paneDropAccepted",
        "paneDeleteButtonClicked",
        "oscReady",
        "pollOSC",
        "oscScriptFinished",
        "contextHelpChanged",
        "inProgressSelectionChanged",
        "openSessionFromRDF",
        "FileOpenStatus",
        "FileSource",
        "source",
        "openLayersFromRDF",
        "closeSession",
        "emitHideSplash",
        "newerVersionAvailable",
        "menuActionMapperInvoked"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'canAddPane'
        QtMocHelpers::SignalData<void(bool)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canDeleteCurrentPane'
        QtMocHelpers::SignalData<void(bool)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canAddLayer'
        QtMocHelpers::SignalData<void(bool)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canImportMoreAudio'
        QtMocHelpers::SignalData<void(bool)>(5, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canReplaceMainAudio'
        QtMocHelpers::SignalData<void(bool)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canImportLayer'
        QtMocHelpers::SignalData<void(bool)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canChangeSessionTemplate'
        QtMocHelpers::SignalData<void(bool)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canExportAudio'
        QtMocHelpers::SignalData<void(bool)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canExportLayer'
        QtMocHelpers::SignalData<void(bool)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canExportImage'
        QtMocHelpers::SignalData<void(bool)>(11, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canRenameLayer'
        QtMocHelpers::SignalData<void(bool)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canEditLayer'
        QtMocHelpers::SignalData<void(bool)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canEditLayerTabular'
        QtMocHelpers::SignalData<void(bool)>(14, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canMeasureLayer'
        QtMocHelpers::SignalData<void(bool)>(15, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSelect'
        QtMocHelpers::SignalData<void(bool)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canClearSelection'
        QtMocHelpers::SignalData<void(bool)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canEditSelection'
        QtMocHelpers::SignalData<void(bool)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canDeleteSelection'
        QtMocHelpers::SignalData<void(bool)>(19, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canPaste'
        QtMocHelpers::SignalData<void(bool)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canInsertInstant'
        QtMocHelpers::SignalData<void(bool)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canInsertInstantsAtBoundaries'
        QtMocHelpers::SignalData<void(bool)>(22, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canInsertItemAtSelection'
        QtMocHelpers::SignalData<void(bool)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canRenumberInstants'
        QtMocHelpers::SignalData<void(bool)>(24, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSubdivideInstants'
        QtMocHelpers::SignalData<void(bool)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canWinnowInstants'
        QtMocHelpers::SignalData<void(bool)>(26, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canDeleteCurrentLayer'
        QtMocHelpers::SignalData<void(bool)>(27, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canZoom'
        QtMocHelpers::SignalData<void(bool)>(28, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canScroll'
        QtMocHelpers::SignalData<void(bool)>(29, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canPlay'
        QtMocHelpers::SignalData<void(bool)>(30, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canRecord'
        QtMocHelpers::SignalData<void(bool)>(31, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canFfwd'
        QtMocHelpers::SignalData<void(bool)>(32, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canRewind'
        QtMocHelpers::SignalData<void(bool)>(33, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canPlaySelection'
        QtMocHelpers::SignalData<void(bool)>(34, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSpeedUpPlayback'
        QtMocHelpers::SignalData<void(bool)>(35, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSlowDownPlayback'
        QtMocHelpers::SignalData<void(bool)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canChangePlaybackSpeed'
        QtMocHelpers::SignalData<void(bool)>(37, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSelectPreviousPane'
        QtMocHelpers::SignalData<void(bool)>(38, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSelectNextPane'
        QtMocHelpers::SignalData<void(bool)>(39, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSelectPreviousLayer'
        QtMocHelpers::SignalData<void(bool)>(40, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSelectNextLayer'
        QtMocHelpers::SignalData<void(bool)>(41, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSave'
        QtMocHelpers::SignalData<void(bool)>(42, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'canSaveAs'
        QtMocHelpers::SignalData<void(bool)>(43, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'hideSplash'
        QtMocHelpers::SignalData<void()>(44, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'hideSplash'
        QtMocHelpers::SignalData<void(QWidget *)>(44, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 45, 2 },
        }}),
        // Signal 'sessionLoaded'
        QtMocHelpers::SignalData<void()>(46, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'audioFileLoaded'
        QtMocHelpers::SignalData<void()>(47, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'replacedDocument'
        QtMocHelpers::SignalData<void()>(48, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'activity'
        QtMocHelpers::SignalData<void(QString)>(49, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'preferenceChanged'
        QtMocHelpers::SlotData<void(PropertyContainer::PropertyName)>(50, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 51, 2 },
        }}),
        // Slot 'resizeConstrained'
        QtMocHelpers::SlotData<void(QSize)>(52, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 53, 2 },
        }}),
        // Slot 'recreateAudioIO'
        QtMocHelpers::SlotData<void()>(54, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'zoomIn'
        QtMocHelpers::SlotData<void()>(55, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'zoomOut'
        QtMocHelpers::SlotData<void()>(56, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'zoomToFit'
        QtMocHelpers::SlotData<void()>(57, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'zoomDefault'
        QtMocHelpers::SlotData<void()>(58, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'scrollLeft'
        QtMocHelpers::SlotData<void()>(59, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'scrollRight'
        QtMocHelpers::SlotData<void()>(60, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'jumpLeft'
        QtMocHelpers::SlotData<void()>(61, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'jumpRight'
        QtMocHelpers::SlotData<void()>(62, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'peekLeft'
        QtMocHelpers::SlotData<void()>(63, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'peekRight'
        QtMocHelpers::SlotData<void()>(64, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'showNoOverlays'
        QtMocHelpers::SlotData<void()>(65, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'showMinimalOverlays'
        QtMocHelpers::SlotData<void()>(66, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'showAllOverlays'
        QtMocHelpers::SlotData<void()>(67, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleTimeRulers'
        QtMocHelpers::SlotData<void()>(68, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleZoomWheels'
        QtMocHelpers::SlotData<void()>(69, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'togglePropertyBoxes'
        QtMocHelpers::SlotData<void()>(70, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleStatusBar'
        QtMocHelpers::SlotData<void()>(71, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'toggleCentreLine'
        QtMocHelpers::SlotData<void()>(72, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'play'
        QtMocHelpers::SlotData<void()>(73, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'ffwd'
        QtMocHelpers::SlotData<void()>(74, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'ffwdEnd'
        QtMocHelpers::SlotData<void()>(75, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'rewind'
        QtMocHelpers::SlotData<void()>(76, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'rewindStart'
        QtMocHelpers::SlotData<void()>(77, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'record'
        QtMocHelpers::SlotData<void()>(78, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(79, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'ffwdSimilar'
        QtMocHelpers::SlotData<void()>(80, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'rewindSimilar'
        QtMocHelpers::SlotData<void()>(81, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'deleteCurrentPane'
        QtMocHelpers::SlotData<void()>(82, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'deleteCurrentLayer'
        QtMocHelpers::SlotData<void()>(83, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'editCurrentLayer'
        QtMocHelpers::SlotData<void()>(84, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'previousPane'
        QtMocHelpers::SlotData<void()>(85, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'nextPane'
        QtMocHelpers::SlotData<void()>(86, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'previousLayer'
        QtMocHelpers::SlotData<void()>(87, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'nextLayer'
        QtMocHelpers::SlotData<void()>(88, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playLoopToggled'
        QtMocHelpers::SlotData<void()>(89, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playSelectionToggled'
        QtMocHelpers::SlotData<void()>(90, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playSoloToggled'
        QtMocHelpers::SlotData<void()>(91, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'audioChannelCountIncreased'
        QtMocHelpers::SlotData<void(int)>(92, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 93 },
        }}),
        // Slot 'sampleRateMismatch'
        QtMocHelpers::SlotData<void(sv_samplerate_t, sv_samplerate_t, bool)>(94, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 95, 2 }, { 0x80000000 | 95, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Slot 'audioOverloadPluginDisabled'
        QtMocHelpers::SlotData<void()>(96, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'playbackFrameChanged'
        QtMocHelpers::SlotData<void(sv_frame_t)>(97, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 98, 2 },
        }}),
        // Slot 'globalCentreFrameChanged'
        QtMocHelpers::SlotData<void(sv_frame_t)>(99, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 98, 2 },
        }}),
        // Slot 'viewCentreFrameChanged'
        QtMocHelpers::SlotData<void(View *, sv_frame_t)>(100, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 101, 2 }, { 0x80000000 | 98, 2 },
        }}),
        // Slot 'viewZoomLevelChanged'
        QtMocHelpers::SlotData<void(View *, ZoomLevel, bool)>(102, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 101, 2 }, { 0x80000000 | 103, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Slot 'monitoringLevelsChanged'
        QtMocHelpers::SlotData<void(float, float)>(104, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Float, 2 }, { QMetaType::Float, 2 },
        }}),
        // Slot 'recordDurationChanged'
        QtMocHelpers::SlotData<void(sv_frame_t, sv_samplerate_t)>(105, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 98, 2 }, { 0x80000000 | 95, 2 },
        }}),
        // Slot 'currentPaneChanged'
        QtMocHelpers::SlotData<void(Pane *)>(106, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 },
        }}),
        // Slot 'currentLayerChanged'
        QtMocHelpers::SlotData<void(Pane *, Layer *)>(108, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 }, { 0x80000000 | 109, 2 },
        }}),
        // Slot 'selectAll'
        QtMocHelpers::SlotData<void()>(110, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'selectToStart'
        QtMocHelpers::SlotData<void()>(111, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'selectToEnd'
        QtMocHelpers::SlotData<void()>(112, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'selectVisible'
        QtMocHelpers::SlotData<void()>(113, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'clearSelection'
        QtMocHelpers::SlotData<void()>(114, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'cut'
        QtMocHelpers::SlotData<void()>(115, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'copy'
        QtMocHelpers::SlotData<void()>(116, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'paste'
        QtMocHelpers::SlotData<void()>(117, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'pasteAtPlaybackPosition'
        QtMocHelpers::SlotData<void()>(118, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'pasteRelative'
        QtMocHelpers::SlotData<void(sv_frame_t)>(119, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 98, 120 },
        }}),
        // Slot 'deleteSelected'
        QtMocHelpers::SlotData<void()>(121, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'insertInstant'
        QtMocHelpers::SlotData<void()>(122, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'insertInstantAt'
        QtMocHelpers::SlotData<void(sv_frame_t)>(123, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 98, 2 },
        }}),
        // Slot 'insertInstantsAtBoundaries'
        QtMocHelpers::SlotData<void()>(124, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'insertItemAtSelection'
        QtMocHelpers::SlotData<void()>(125, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'insertItemAt'
        QtMocHelpers::SlotData<void(sv_frame_t, sv_frame_t)>(126, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 98, 2 }, { 0x80000000 | 98, 2 },
        }}),
        // Slot 'renumberInstants'
        QtMocHelpers::SlotData<void()>(127, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'subdivideInstantsBy'
        QtMocHelpers::SlotData<void(int)>(128, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'winnowInstantsBy'
        QtMocHelpers::SlotData<void(int)>(129, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::Int, 2 },
        }}),
        // Slot 'documentModified'
        QtMocHelpers::SlotData<void()>(130, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'documentRestored'
        QtMocHelpers::SlotData<void()>(131, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'layerAdded'
        QtMocHelpers::SlotData<void(Layer *)>(132, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 109, 2 },
        }}),
        // Slot 'layerRemoved'
        QtMocHelpers::SlotData<void(Layer *)>(133, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 109, 2 },
        }}),
        // Slot 'layerAboutToBeDeleted'
        QtMocHelpers::SlotData<void(Layer *)>(134, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 109, 2 },
        }}),
        // Slot 'layerInAView'
        QtMocHelpers::SlotData<void(Layer *, bool)>(135, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 109, 2 }, { QMetaType::Bool, 2 },
        }}),
        // Slot 'mainModelChanged'
        QtMocHelpers::SlotData<void(ModelId)>(136, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 137, 2 },
        }}),
        // Slot 'modelAdded'
        QtMocHelpers::SlotData<void(ModelId)>(138, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 137, 2 },
        }}),
        // Slot 'updateMenuStates'
        QtMocHelpers::SlotData<void()>(139, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'updateDescriptionLabel'
        QtMocHelpers::SlotData<void()>(140, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'updateWindowTitle'
        QtMocHelpers::SlotData<void()>(141, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'modelGenerationFailed'
        QtMocHelpers::SlotData<void(QString, QString)>(142, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'modelGenerationWarning'
        QtMocHelpers::SlotData<void(QString, QString)>(143, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'modelRegenerationFailed'
        QtMocHelpers::SlotData<void(QString, QString, QString)>(144, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'modelRegenerationWarning'
        QtMocHelpers::SlotData<void(QString, QString, QString)>(145, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 }, { QMetaType::QString, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'alignmentComplete'
        QtMocHelpers::SlotData<void(ModelId)>(146, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 137, 2 },
        }}),
        // Slot 'alignmentFailed'
        QtMocHelpers::SlotData<void(ModelId, QString)>(147, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 137, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'paneRightButtonMenuRequested'
        QtMocHelpers::SlotData<void(Pane *, QPoint)>(148, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 }, { 0x80000000 | 149, 150 },
        }}),
        // Slot 'panePropertiesRightButtonMenuRequested'
        QtMocHelpers::SlotData<void(Pane *, QPoint)>(151, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 }, { 0x80000000 | 149, 150 },
        }}),
        // Slot 'layerPropertiesRightButtonMenuRequested'
        QtMocHelpers::SlotData<void(Pane *, Layer *, QPoint)>(152, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 }, { 0x80000000 | 109, 2 }, { 0x80000000 | 149, 150 },
        }}),
        // Slot 'paneAdded'
        QtMocHelpers::SlotData<void(Pane *)>(153, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 },
        }}),
        // Slot 'paneHidden'
        QtMocHelpers::SlotData<void(Pane *)>(154, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 },
        }}),
        // Slot 'paneAboutToBeDeleted'
        QtMocHelpers::SlotData<void(Pane *)>(155, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 },
        }}),
        // Slot 'paneDropAccepted'
        QtMocHelpers::SlotData<void(Pane *, QStringList)>(156, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 }, { QMetaType::QStringList, 2 },
        }}),
        // Slot 'paneDropAccepted'
        QtMocHelpers::SlotData<void(Pane *, QString)>(156, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 }, { QMetaType::QString, 2 },
        }}),
        // Slot 'paneDeleteButtonClicked'
        QtMocHelpers::SlotData<void(Pane *)>(157, 2, QMC::AccessProtected, QMetaType::Void, {{
            { 0x80000000 | 107, 2 },
        }}),
        // Slot 'oscReady'
        QtMocHelpers::SlotData<void()>(158, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'pollOSC'
        QtMocHelpers::SlotData<void()>(159, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'oscScriptFinished'
        QtMocHelpers::SlotData<void()>(160, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'contextHelpChanged'
        QtMocHelpers::SlotData<void(const QString &)>(161, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'inProgressSelectionChanged'
        QtMocHelpers::SlotData<void()>(162, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'openSessionFromRDF'
        QtMocHelpers::SlotData<enum FileOpenStatus(FileSource)>(163, 2, QMC::AccessProtected, 0x80000000 | 164, {{
            { 0x80000000 | 165, 166 },
        }}),
        // Slot 'openLayersFromRDF'
        QtMocHelpers::SlotData<enum FileOpenStatus(FileSource)>(167, 2, QMC::AccessProtected, 0x80000000 | 164, {{
            { 0x80000000 | 165, 166 },
        }}),
        // Slot 'closeSession'
        QtMocHelpers::SlotData<void()>(168, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'emitHideSplash'
        QtMocHelpers::SlotData<void()>(169, 2, QMC::AccessProtected, QMetaType::Void),
        // Slot 'newerVersionAvailable'
        QtMocHelpers::SlotData<void(QString)>(170, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QString, 2 },
        }}),
        // Slot 'menuActionMapperInvoked'
        QtMocHelpers::SlotData<void(QObject *)>(171, 2, QMC::AccessProtected, QMetaType::Void, {{
            { QMetaType::QObjectStar, 2 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<MainWindowBase, qt_meta_tag_ZN2sv14MainWindowBaseE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::MainWindowBase::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14MainWindowBaseE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14MainWindowBaseE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv14MainWindowBaseE_t>.metaTypes,
    nullptr
} };

void sv::MainWindowBase::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<MainWindowBase *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->canAddPane((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 1: _t->canDeleteCurrentPane((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 2: _t->canAddLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 3: _t->canImportMoreAudio((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 4: _t->canReplaceMainAudio((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 5: _t->canImportLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 6: _t->canChangeSessionTemplate((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 7: _t->canExportAudio((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 8: _t->canExportLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 9: _t->canExportImage((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 10: _t->canRenameLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 11: _t->canEditLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 12: _t->canEditLayerTabular((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 13: _t->canMeasureLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 14: _t->canSelect((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 15: _t->canClearSelection((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 16: _t->canEditSelection((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 17: _t->canDeleteSelection((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 18: _t->canPaste((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 19: _t->canInsertInstant((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 20: _t->canInsertInstantsAtBoundaries((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 21: _t->canInsertItemAtSelection((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 22: _t->canRenumberInstants((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 23: _t->canSubdivideInstants((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 24: _t->canWinnowInstants((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 25: _t->canDeleteCurrentLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 26: _t->canZoom((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 27: _t->canScroll((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 28: _t->canPlay((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 29: _t->canRecord((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 30: _t->canFfwd((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 31: _t->canRewind((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 32: _t->canPlaySelection((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 33: _t->canSpeedUpPlayback((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 34: _t->canSlowDownPlayback((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 35: _t->canChangePlaybackSpeed((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 36: _t->canSelectPreviousPane((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 37: _t->canSelectNextPane((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 38: _t->canSelectPreviousLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 39: _t->canSelectNextLayer((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 40: _t->canSave((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 41: _t->canSaveAs((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 42: _t->hideSplash(); break;
        case 43: _t->hideSplash((*reinterpret_cast<std::add_pointer_t<QWidget*>>(_a[1]))); break;
        case 44: _t->sessionLoaded(); break;
        case 45: _t->audioFileLoaded(); break;
        case 46: _t->replacedDocument(); break;
        case 47: _t->activity((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 48: _t->preferenceChanged((*reinterpret_cast<std::add_pointer_t<PropertyContainer::PropertyName>>(_a[1]))); break;
        case 49: _t->resizeConstrained((*reinterpret_cast<std::add_pointer_t<QSize>>(_a[1]))); break;
        case 50: _t->recreateAudioIO(); break;
        case 51: _t->zoomIn(); break;
        case 52: _t->zoomOut(); break;
        case 53: _t->zoomToFit(); break;
        case 54: _t->zoomDefault(); break;
        case 55: _t->scrollLeft(); break;
        case 56: _t->scrollRight(); break;
        case 57: _t->jumpLeft(); break;
        case 58: _t->jumpRight(); break;
        case 59: _t->peekLeft(); break;
        case 60: _t->peekRight(); break;
        case 61: _t->showNoOverlays(); break;
        case 62: _t->showMinimalOverlays(); break;
        case 63: _t->showAllOverlays(); break;
        case 64: _t->toggleTimeRulers(); break;
        case 65: _t->toggleZoomWheels(); break;
        case 66: _t->togglePropertyBoxes(); break;
        case 67: _t->toggleStatusBar(); break;
        case 68: _t->toggleCentreLine(); break;
        case 69: _t->play(); break;
        case 70: _t->ffwd(); break;
        case 71: _t->ffwdEnd(); break;
        case 72: _t->rewind(); break;
        case 73: _t->rewindStart(); break;
        case 74: _t->record(); break;
        case 75: _t->stop(); break;
        case 76: _t->ffwdSimilar(); break;
        case 77: _t->rewindSimilar(); break;
        case 78: _t->deleteCurrentPane(); break;
        case 79: _t->deleteCurrentLayer(); break;
        case 80: _t->editCurrentLayer(); break;
        case 81: _t->previousPane(); break;
        case 82: _t->nextPane(); break;
        case 83: _t->previousLayer(); break;
        case 84: _t->nextLayer(); break;
        case 85: _t->playLoopToggled(); break;
        case 86: _t->playSelectionToggled(); break;
        case 87: _t->playSoloToggled(); break;
        case 88: _t->audioChannelCountIncreased((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 89: _t->sampleRateMismatch((*reinterpret_cast<std::add_pointer_t<sv_samplerate_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_samplerate_t>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 90: _t->audioOverloadPluginDisabled(); break;
        case 91: _t->playbackFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 92: _t->globalCentreFrameChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 93: _t->viewCentreFrameChanged((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2]))); break;
        case 94: _t->viewZoomLevelChanged((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<ZoomLevel>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 95: _t->monitoringLevelsChanged((*reinterpret_cast<std::add_pointer_t<float>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[2]))); break;
        case 96: _t->recordDurationChanged((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_samplerate_t>>(_a[2]))); break;
        case 97: _t->currentPaneChanged((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 98: _t->currentLayerChanged((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[2]))); break;
        case 99: _t->selectAll(); break;
        case 100: _t->selectToStart(); break;
        case 101: _t->selectToEnd(); break;
        case 102: _t->selectVisible(); break;
        case 103: _t->clearSelection(); break;
        case 104: _t->cut(); break;
        case 105: _t->copy(); break;
        case 106: _t->paste(); break;
        case 107: _t->pasteAtPlaybackPosition(); break;
        case 108: _t->pasteRelative((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 109: _t->deleteSelected(); break;
        case 110: _t->insertInstant(); break;
        case 111: _t->insertInstantAt((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 112: _t->insertInstantsAtBoundaries(); break;
        case 113: _t->insertItemAtSelection(); break;
        case 114: _t->insertItemAt((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[2]))); break;
        case 115: _t->renumberInstants(); break;
        case 116: _t->subdivideInstantsBy((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 117: _t->winnowInstantsBy((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 118: _t->documentModified(); break;
        case 119: _t->documentRestored(); break;
        case 120: _t->layerAdded((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1]))); break;
        case 121: _t->layerRemoved((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1]))); break;
        case 122: _t->layerAboutToBeDeleted((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1]))); break;
        case 123: _t->layerInAView((*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 124: _t->mainModelChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 125: _t->modelAdded((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 126: _t->updateMenuStates(); break;
        case 127: _t->updateDescriptionLabel(); break;
        case 128: _t->updateWindowTitle(); break;
        case 129: _t->modelGenerationFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 130: _t->modelGenerationWarning((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 131: _t->modelRegenerationFailed((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 132: _t->modelRegenerationWarning((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[3]))); break;
        case 133: _t->alignmentComplete((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 134: _t->alignmentFailed((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 135: _t->paneRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[2]))); break;
        case 136: _t->panePropertiesRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[2]))); break;
        case 137: _t->layerPropertiesRightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<Layer*>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[3]))); break;
        case 138: _t->paneAdded((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 139: _t->paneHidden((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 140: _t->paneAboutToBeDeleted((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 141: _t->paneDropAccepted((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[2]))); break;
        case 142: _t->paneDropAccepted((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 143: _t->paneDeleteButtonClicked((*reinterpret_cast<std::add_pointer_t<Pane*>>(_a[1]))); break;
        case 144: _t->oscReady(); break;
        case 145: _t->pollOSC(); break;
        case 146: _t->oscScriptFinished(); break;
        case 147: _t->contextHelpChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 148: _t->inProgressSelectionChanged(); break;
        case 149: { enum FileOpenStatus _r = _t->openSessionFromRDF((*reinterpret_cast<std::add_pointer_t<FileSource>>(_a[1])));
            if (_a[0]) *reinterpret_cast<enum FileOpenStatus*>(_a[0]) = std::move(_r); }  break;
        case 150: { enum FileOpenStatus _r = _t->openLayersFromRDF((*reinterpret_cast<std::add_pointer_t<FileSource>>(_a[1])));
            if (_a[0]) *reinterpret_cast<enum FileOpenStatus*>(_a[0]) = std::move(_r); }  break;
        case 151: _t->closeSession(); break;
        case 152: _t->emitHideSplash(); break;
        case 153: _t->newerVersionAvailable((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 154: _t->menuActionMapperInvoked((*reinterpret_cast<std::add_pointer_t<QObject*>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canAddPane, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canDeleteCurrentPane, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canAddLayer, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canImportMoreAudio, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canReplaceMainAudio, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canImportLayer, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canChangeSessionTemplate, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canExportAudio, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canExportLayer, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canExportImage, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canRenameLayer, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canEditLayer, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canEditLayerTabular, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canMeasureLayer, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSelect, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canClearSelection, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canEditSelection, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canDeleteSelection, 17))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canPaste, 18))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canInsertInstant, 19))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canInsertInstantsAtBoundaries, 20))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canInsertItemAtSelection, 21))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canRenumberInstants, 22))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSubdivideInstants, 23))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canWinnowInstants, 24))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canDeleteCurrentLayer, 25))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canZoom, 26))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canScroll, 27))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canPlay, 28))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canRecord, 29))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canFfwd, 30))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canRewind, 31))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canPlaySelection, 32))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSpeedUpPlayback, 33))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSlowDownPlayback, 34))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canChangePlaybackSpeed, 35))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSelectPreviousPane, 36))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSelectNextPane, 37))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSelectPreviousLayer, 38))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSelectNextLayer, 39))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSave, 40))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(bool )>(_a, &MainWindowBase::canSaveAs, 41))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)()>(_a, &MainWindowBase::hideSplash, 42))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(QWidget * )>(_a, &MainWindowBase::hideSplash, 43))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)()>(_a, &MainWindowBase::sessionLoaded, 44))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)()>(_a, &MainWindowBase::audioFileLoaded, 45))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)()>(_a, &MainWindowBase::replacedDocument, 46))
            return;
        if (QtMocHelpers::indexOfMethod<void (MainWindowBase::*)(QString )>(_a, &MainWindowBase::activity, 47))
            return;
    }
}

const QMetaObject *sv::MainWindowBase::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::MainWindowBase::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv14MainWindowBaseE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "FrameTimer"))
        return static_cast< FrameTimer*>(this);
    if (!strcmp(_clname, "OSCMessageCallback"))
        return static_cast< OSCMessageCallback*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int sv::MainWindowBase::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 155)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 155;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 155)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 155;
    }
    return _id;
}

// SIGNAL 0
void sv::MainWindowBase::canAddPane(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void sv::MainWindowBase::canDeleteCurrentPane(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void sv::MainWindowBase::canAddLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::MainWindowBase::canImportMoreAudio(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void sv::MainWindowBase::canReplaceMainAudio(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void sv::MainWindowBase::canImportLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void sv::MainWindowBase::canChangeSessionTemplate(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void sv::MainWindowBase::canExportAudio(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void sv::MainWindowBase::canExportLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}

// SIGNAL 9
void sv::MainWindowBase::canExportImage(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 9, nullptr, _t1);
}

// SIGNAL 10
void sv::MainWindowBase::canRenameLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 10, nullptr, _t1);
}

// SIGNAL 11
void sv::MainWindowBase::canEditLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 11, nullptr, _t1);
}

// SIGNAL 12
void sv::MainWindowBase::canEditLayerTabular(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 12, nullptr, _t1);
}

// SIGNAL 13
void sv::MainWindowBase::canMeasureLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 13, nullptr, _t1);
}

// SIGNAL 14
void sv::MainWindowBase::canSelect(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 14, nullptr, _t1);
}

// SIGNAL 15
void sv::MainWindowBase::canClearSelection(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1);
}

// SIGNAL 16
void sv::MainWindowBase::canEditSelection(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 16, nullptr, _t1);
}

// SIGNAL 17
void sv::MainWindowBase::canDeleteSelection(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 17, nullptr, _t1);
}

// SIGNAL 18
void sv::MainWindowBase::canPaste(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 18, nullptr, _t1);
}

// SIGNAL 19
void sv::MainWindowBase::canInsertInstant(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 19, nullptr, _t1);
}

// SIGNAL 20
void sv::MainWindowBase::canInsertInstantsAtBoundaries(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 20, nullptr, _t1);
}

// SIGNAL 21
void sv::MainWindowBase::canInsertItemAtSelection(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 21, nullptr, _t1);
}

// SIGNAL 22
void sv::MainWindowBase::canRenumberInstants(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 22, nullptr, _t1);
}

// SIGNAL 23
void sv::MainWindowBase::canSubdivideInstants(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 23, nullptr, _t1);
}

// SIGNAL 24
void sv::MainWindowBase::canWinnowInstants(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 24, nullptr, _t1);
}

// SIGNAL 25
void sv::MainWindowBase::canDeleteCurrentLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 25, nullptr, _t1);
}

// SIGNAL 26
void sv::MainWindowBase::canZoom(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 26, nullptr, _t1);
}

// SIGNAL 27
void sv::MainWindowBase::canScroll(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 27, nullptr, _t1);
}

// SIGNAL 28
void sv::MainWindowBase::canPlay(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 28, nullptr, _t1);
}

// SIGNAL 29
void sv::MainWindowBase::canRecord(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 29, nullptr, _t1);
}

// SIGNAL 30
void sv::MainWindowBase::canFfwd(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 30, nullptr, _t1);
}

// SIGNAL 31
void sv::MainWindowBase::canRewind(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 31, nullptr, _t1);
}

// SIGNAL 32
void sv::MainWindowBase::canPlaySelection(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 32, nullptr, _t1);
}

// SIGNAL 33
void sv::MainWindowBase::canSpeedUpPlayback(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 33, nullptr, _t1);
}

// SIGNAL 34
void sv::MainWindowBase::canSlowDownPlayback(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 34, nullptr, _t1);
}

// SIGNAL 35
void sv::MainWindowBase::canChangePlaybackSpeed(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 35, nullptr, _t1);
}

// SIGNAL 36
void sv::MainWindowBase::canSelectPreviousPane(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 36, nullptr, _t1);
}

// SIGNAL 37
void sv::MainWindowBase::canSelectNextPane(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 37, nullptr, _t1);
}

// SIGNAL 38
void sv::MainWindowBase::canSelectPreviousLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 38, nullptr, _t1);
}

// SIGNAL 39
void sv::MainWindowBase::canSelectNextLayer(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 39, nullptr, _t1);
}

// SIGNAL 40
void sv::MainWindowBase::canSave(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 40, nullptr, _t1);
}

// SIGNAL 41
void sv::MainWindowBase::canSaveAs(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 41, nullptr, _t1);
}

// SIGNAL 42
void sv::MainWindowBase::hideSplash()
{
    QMetaObject::activate(this, &staticMetaObject, 42, nullptr);
}

// SIGNAL 43
void sv::MainWindowBase::hideSplash(QWidget * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 43, nullptr, _t1);
}

// SIGNAL 44
void sv::MainWindowBase::sessionLoaded()
{
    QMetaObject::activate(this, &staticMetaObject, 44, nullptr);
}

// SIGNAL 45
void sv::MainWindowBase::audioFileLoaded()
{
    QMetaObject::activate(this, &staticMetaObject, 45, nullptr);
}

// SIGNAL 46
void sv::MainWindowBase::replacedDocument()
{
    QMetaObject::activate(this, &staticMetaObject, 46, nullptr);
}

// SIGNAL 47
void sv::MainWindowBase::activity(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 47, nullptr, _t1);
}
QT_WARNING_POP
