/****************************************************************************
** Meta object code from reading C++ file 'Pane.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../sonic-visualiser-5.2.1/svgui/view/Pane.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'Pane.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN2sv4PaneE_t {};
} // unnamed namespace

template <> constexpr inline auto sv::Pane::qt_create_metaobjectdata<qt_meta_tag_ZN2sv4PaneE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "sv::Pane",
        "paneInteractedWith",
        "",
        "rightButtonMenuRequested",
        "QPoint",
        "position",
        "dropAccepted",
        "uriList",
        "text",
        "doubleClickSelectInvoked",
        "sv_frame_t",
        "frame",
        "regionOutlined",
        "QRect",
        "rect",
        "toolModeChanged",
        "zoomWheelsEnabledChanged",
        "viewZoomLevelChanged",
        "View*",
        "ZoomLevel",
        "locked",
        "modelAlignmentCompletionChanged",
        "ModelId",
        "horizontalThumbwheelMoved",
        "value",
        "verticalThumbwheelMoved",
        "verticalZoomChanged",
        "verticalPannerMoved",
        "x",
        "y",
        "w",
        "h",
        "verticalPannerContextMenuRequested",
        "editVerticalPannerExtents",
        "resetVerticalPannerExtents",
        "layerParametersChanged",
        "propertyContainerSelected",
        "PropertyContainer*",
        "pc",
        "zoomToRegion",
        "r",
        "mouseEnteredWidget",
        "mouseLeftWidget",
        "playbackScheduleTimerElapsed"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'paneInteractedWith'
        QtMocHelpers::SignalData<void()>(1, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'rightButtonMenuRequested'
        QtMocHelpers::SignalData<void(QPoint)>(3, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 5 },
        }}),
        // Signal 'dropAccepted'
        QtMocHelpers::SignalData<void(QStringList)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QStringList, 7 },
        }}),
        // Signal 'dropAccepted'
        QtMocHelpers::SignalData<void(QString)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 8 },
        }}),
        // Signal 'doubleClickSelectInvoked'
        QtMocHelpers::SignalData<void(sv_frame_t)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 10, 11 },
        }}),
        // Signal 'regionOutlined'
        QtMocHelpers::SignalData<void(QRect)>(12, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 14 },
        }}),
        // Slot 'toolModeChanged'
        QtMocHelpers::SlotData<void()>(15, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'zoomWheelsEnabledChanged'
        QtMocHelpers::SlotData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'viewZoomLevelChanged'
        QtMocHelpers::SlotData<void(View *, ZoomLevel, bool)>(17, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 18, 2 }, { 0x80000000 | 19, 2 }, { QMetaType::Bool, 20 },
        }}),
        // Slot 'modelAlignmentCompletionChanged'
        QtMocHelpers::SlotData<void(ModelId)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 22, 2 },
        }}),
        // Slot 'horizontalThumbwheelMoved'
        QtMocHelpers::SlotData<void(int)>(23, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 24 },
        }}),
        // Slot 'verticalThumbwheelMoved'
        QtMocHelpers::SlotData<void(int)>(25, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 24 },
        }}),
        // Slot 'verticalZoomChanged'
        QtMocHelpers::SlotData<void()>(26, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'verticalPannerMoved'
        QtMocHelpers::SlotData<void(float, float, float, float)>(27, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 28 }, { QMetaType::Float, 29 }, { QMetaType::Float, 30 }, { QMetaType::Float, 31 },
        }}),
        // Slot 'verticalPannerContextMenuRequested'
        QtMocHelpers::SlotData<void(const QPoint &)>(32, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 4, 2 },
        }}),
        // Slot 'editVerticalPannerExtents'
        QtMocHelpers::SlotData<void()>(33, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'resetVerticalPannerExtents'
        QtMocHelpers::SlotData<void()>(34, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'layerParametersChanged'
        QtMocHelpers::SlotData<void()>(35, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'propertyContainerSelected'
        QtMocHelpers::SlotData<void(View *, PropertyContainer *)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 18, 2 }, { 0x80000000 | 37, 38 },
        }}),
        // Slot 'zoomToRegion'
        QtMocHelpers::SlotData<void(QRect)>(39, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 13, 40 },
        }}),
        // Slot 'mouseEnteredWidget'
        QtMocHelpers::SlotData<void()>(41, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'mouseLeftWidget'
        QtMocHelpers::SlotData<void()>(42, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'playbackScheduleTimerElapsed'
        QtMocHelpers::SlotData<void()>(43, 2, QMC::AccessProtected, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<Pane, qt_meta_tag_ZN2sv4PaneE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject sv::Pane::staticMetaObject = { {
    QMetaObject::SuperData::link<View::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv4PaneE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv4PaneE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN2sv4PaneE_t>.metaTypes,
    nullptr
} };

void sv::Pane::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<Pane *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->paneInteractedWith(); break;
        case 1: _t->rightButtonMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 2: _t->dropAccepted((*reinterpret_cast<std::add_pointer_t<QStringList>>(_a[1]))); break;
        case 3: _t->dropAccepted((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 4: _t->doubleClickSelectInvoked((*reinterpret_cast<std::add_pointer_t<sv_frame_t>>(_a[1]))); break;
        case 5: _t->regionOutlined((*reinterpret_cast<std::add_pointer_t<QRect>>(_a[1]))); break;
        case 6: _t->toolModeChanged(); break;
        case 7: _t->zoomWheelsEnabledChanged(); break;
        case 8: _t->viewZoomLevelChanged((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<ZoomLevel>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[3]))); break;
        case 9: _t->modelAlignmentCompletionChanged((*reinterpret_cast<std::add_pointer_t<ModelId>>(_a[1]))); break;
        case 10: _t->horizontalThumbwheelMoved((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 11: _t->verticalThumbwheelMoved((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 12: _t->verticalZoomChanged(); break;
        case 13: _t->verticalPannerMoved((*reinterpret_cast<std::add_pointer_t<float>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[2])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[3])),(*reinterpret_cast<std::add_pointer_t<float>>(_a[4]))); break;
        case 14: _t->verticalPannerContextMenuRequested((*reinterpret_cast<std::add_pointer_t<QPoint>>(_a[1]))); break;
        case 15: _t->editVerticalPannerExtents(); break;
        case 16: _t->resetVerticalPannerExtents(); break;
        case 17: _t->layerParametersChanged(); break;
        case 18: _t->propertyContainerSelected((*reinterpret_cast<std::add_pointer_t<View*>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<PropertyContainer*>>(_a[2]))); break;
        case 19: _t->zoomToRegion((*reinterpret_cast<std::add_pointer_t<QRect>>(_a[1]))); break;
        case 20: _t->mouseEnteredWidget(); break;
        case 21: _t->mouseLeftWidget(); break;
        case 22: _t->playbackScheduleTimerElapsed(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 8:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< View* >(); break;
            }
            break;
        case 18:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< View* >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (Pane::*)()>(_a, &Pane::paneInteractedWith, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (Pane::*)(QPoint )>(_a, &Pane::rightButtonMenuRequested, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (Pane::*)(QStringList )>(_a, &Pane::dropAccepted, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (Pane::*)(QString )>(_a, &Pane::dropAccepted, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (Pane::*)(sv_frame_t )>(_a, &Pane::doubleClickSelectInvoked, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (Pane::*)(QRect )>(_a, &Pane::regionOutlined, 5))
            return;
    }
}

const QMetaObject *sv::Pane::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *sv::Pane::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN2sv4PaneE_t>.strings))
        return static_cast<void*>(this);
    return View::qt_metacast(_clname);
}

int sv::Pane::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = View::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 23)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 23;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 23)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 23;
    }
    return _id;
}

// SIGNAL 0
void sv::Pane::paneInteractedWith()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void sv::Pane::rightButtonMenuRequested(QPoint _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void sv::Pane::dropAccepted(QStringList _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void sv::Pane::dropAccepted(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void sv::Pane::doubleClickSelectInvoked(sv_frame_t _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void sv::Pane::regionOutlined(QRect _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}
QT_WARNING_POP
